import json
from typing import Any, Dict, Union

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse
from moto.utilities.aws_headers import amzn_request_id
from .models import databrew_backends, DataBrewBackend


class DataBrewResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="databrew")

    @property
    def databrew_backend(self) -> DataBrewBackend:
        """Return backend instance specific for this region."""
        return databrew_backends[self.current_account][self.region]

    # region Recipes
    @property
    def parameters(self) -> Dict[str, Any]:  # type: ignore[misc]
        return json.loads(self.body)

    @amzn_request_id
    def create_recipe(self) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_CreateRecipe.html
        recipe_description = self.parameters.get("Description")
        recipe_steps = self.parameters.get("Steps")
        recipe_name = self.parameters.get("Name")
        tags = self.parameters.get("Tags")
        return json.dumps(
            self.databrew_backend.create_recipe(
                recipe_name, recipe_description, recipe_steps, tags  # type: ignore[arg-type]
            ).as_dict()
        )

    @amzn_request_id
    def delete_recipe_version(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        # https://docs.aws.amazon.com/databrew/latest/dg/API_DeleteRecipeVersion.html
        if request.method == "DELETE":
            split_path = self.parsed_url.path.strip("/").split("/")
            recipe_name = split_path[1]
            recipe_version = split_path[3]
            self.databrew_backend.delete_recipe_version(recipe_name, recipe_version)
            return (
                200,
                {},
                json.dumps({"Name": recipe_name, "RecipeVersion": recipe_version}),
            )

    @amzn_request_id
    def list_recipes(self) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_ListRecipes.html
        next_token = self._get_param("NextToken", self._get_param("nextToken"))
        max_results = self._get_int_param(
            "MaxResults", self._get_int_param("maxResults")
        )
        recipe_version = self._get_param(
            "RecipeVersion", self._get_param("recipeVersion")
        )

        # pylint: disable=unexpected-keyword-arg, unbalanced-tuple-unpacking
        recipe_list, next_token = self.databrew_backend.list_recipes(
            next_token=next_token,
            max_results=max_results,
            recipe_version=recipe_version,
        )
        return json.dumps(
            {
                "Recipes": [recipe.as_dict() for recipe in recipe_list],
                "NextToken": next_token,
            }
        )

    @amzn_request_id
    def list_recipe_versions(self, request: Any, full_url: str, headers: Any) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_ListRecipeVersions.html
        self.setup_class(request, full_url, headers)
        recipe_name = self._get_param("Name", self._get_param("name"))
        next_token = self._get_param("NextToken", self._get_param("nextToken"))
        max_results = self._get_int_param(
            "MaxResults", self._get_int_param("maxResults")
        )

        # pylint: disable=unexpected-keyword-arg, unbalanced-tuple-unpacking
        recipe_list, next_token = self.databrew_backend.list_recipe_versions(
            recipe_name=recipe_name, next_token=next_token, max_results=max_results
        )
        return json.dumps(
            {
                "Recipes": [recipe.as_dict() for recipe in recipe_list],
                "NextToken": next_token,
            }
        )

    @amzn_request_id
    def publish_recipe(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "POST":
            recipe_name = self.parsed_url.path.strip("/").split("/", 2)[1]
            recipe_description = self.parameters.get("Description")
            self.databrew_backend.publish_recipe(recipe_name, recipe_description)
            return 200, {}, json.dumps({"Name": recipe_name})

    def put_recipe_response(self, recipe_name: str) -> TYPE_RESPONSE:
        recipe_description = self.parameters.get("Description")
        recipe_steps = self.parameters.get("Steps")

        self.databrew_backend.update_recipe(
            recipe_name, recipe_description, recipe_steps  # type: ignore[arg-type]
        )
        return 200, {}, json.dumps({"Name": recipe_name})

    def get_recipe_response(self, recipe_name: str) -> TYPE_RESPONSE:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_DescribeRecipe.html
        recipe_version = self._get_param(
            "RecipeVersion", self._get_param("recipeVersion")
        )
        recipe = self.databrew_backend.describe_recipe(
            recipe_name, recipe_version=recipe_version
        )
        return 200, {}, json.dumps(recipe.as_dict())

    @amzn_request_id
    def recipe_response(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)

        recipe_name = self.parsed_url.path.rstrip("/").rsplit("/", 1)[1]

        if request.method == "PUT":
            return self.put_recipe_response(recipe_name)
        elif request.method == "GET":
            return self.get_recipe_response(recipe_name)

    # endregion

    # region Rulesets

    @amzn_request_id
    def create_ruleset(self) -> str:
        ruleset_description = self.parameters.get("Description")
        ruleset_rules = self.parameters.get("Rules")
        ruleset_name = self.parameters.get("Name")
        ruleset_target_arn = self.parameters.get("TargetArn")
        tags = self.parameters.get("Tags")

        return json.dumps(
            self.databrew_backend.create_ruleset(
                ruleset_name,  # type: ignore[arg-type]
                ruleset_description,  # type: ignore[arg-type]
                ruleset_rules,  # type: ignore[arg-type]
                ruleset_target_arn,  # type: ignore[arg-type]
                tags,  # type: ignore[arg-type]
            ).as_dict()
        )

    def put_ruleset_response(self, ruleset_name: str) -> TYPE_RESPONSE:
        ruleset_description = self.parameters.get("Description")
        ruleset_rules = self.parameters.get("Rules")
        tags = self.parameters.get("Tags")

        ruleset = self.databrew_backend.update_ruleset(
            ruleset_name, ruleset_description, ruleset_rules, tags  # type: ignore[arg-type]
        )
        return 200, {}, json.dumps(ruleset.as_dict())

    def get_ruleset_response(self, ruleset_name: str) -> TYPE_RESPONSE:
        ruleset = self.databrew_backend.describe_ruleset(ruleset_name)
        return 200, {}, json.dumps(ruleset.as_dict())

    def delete_ruleset_response(self, ruleset_name: str) -> TYPE_RESPONSE:
        self.databrew_backend.delete_ruleset(ruleset_name)
        return 200, {}, json.dumps({"Name": ruleset_name})

    @amzn_request_id
    def ruleset_response(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)

        ruleset_name = self.parsed_url.path.split("/")[-1]

        if request.method == "PUT":
            response = self.put_ruleset_response(ruleset_name)
            return response
        elif request.method == "GET":
            return self.get_ruleset_response(ruleset_name)
        elif request.method == "DELETE":
            return self.delete_ruleset_response(ruleset_name)

    @amzn_request_id
    def list_rulesets(self) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_ListRulesets.html
        next_token = self._get_param("NextToken", self._get_param("nextToken"))
        max_results = self._get_int_param(
            "MaxResults", self._get_int_param("maxResults")
        )

        # pylint: disable=unexpected-keyword-arg, unbalanced-tuple-unpacking
        ruleset_list, next_token = self.databrew_backend.list_rulesets(
            next_token=next_token, max_results=max_results
        )
        return json.dumps(
            {
                "Rulesets": [ruleset.as_dict() for ruleset in ruleset_list],
                "NextToken": next_token,
            }
        )

    # endregion

    # region Datasets

    @amzn_request_id
    def create_dataset(self) -> str:
        dataset_name = self.parameters.get("Name")
        dataset_format = self.parameters.get("Format")
        dataset_format_options = self.parameters.get("FormatOptions")
        dataset_input = self.parameters.get("Input")
        dataset_path_otions = self.parameters.get("PathOptions")
        dataset_tags = self.parameters.get("Tags")

        return json.dumps(
            self.databrew_backend.create_dataset(
                dataset_name,  # type: ignore[arg-type]
                dataset_format,  # type: ignore[arg-type]
                dataset_format_options,  # type: ignore[arg-type]
                dataset_input,  # type: ignore[arg-type]
                dataset_path_otions,  # type: ignore[arg-type]
                dataset_tags,  # type: ignore[arg-type]
            ).as_dict()
        )

    @amzn_request_id
    def list_datasets(self) -> str:
        next_token = self._get_param("NextToken", self._get_param("nextToken"))
        max_results = self._get_int_param(
            "MaxResults", self._get_int_param("maxResults")
        )

        # pylint: disable=unexpected-keyword-arg, unbalanced-tuple-unpacking
        dataset_list, next_token = self.databrew_backend.list_datasets(
            next_token=next_token, max_results=max_results
        )

        return json.dumps(
            {
                "Datasets": [dataset.as_dict() for dataset in dataset_list],
                "NextToken": next_token,
            }
        )

    @amzn_request_id
    def update_dataset(self, dataset_name: str) -> TYPE_RESPONSE:
        dataset_format = self.parameters.get("Format")
        dataset_format_options = self.parameters.get("FormatOptions")
        dataset_input = self.parameters.get("Input")
        dataset_path_otions = self.parameters.get("PathOptions")
        dataset_tags = self.parameters.get("Tags")

        dataset = self.databrew_backend.update_dataset(
            dataset_name,
            dataset_format,  # type: ignore[arg-type]
            dataset_format_options,  # type: ignore[arg-type]
            dataset_input,  # type: ignore[arg-type]
            dataset_path_otions,  # type: ignore[arg-type]
            dataset_tags,  # type: ignore[arg-type]
        )
        return 200, {}, json.dumps(dataset.as_dict())

    @amzn_request_id
    def delete_dataset(self, dataset_name: str) -> TYPE_RESPONSE:
        self.databrew_backend.delete_dataset(dataset_name)
        return 200, {}, json.dumps({"Name": dataset_name})

    @amzn_request_id
    def describe_dataset(self, dataset_name: str) -> TYPE_RESPONSE:
        dataset = self.databrew_backend.describe_dataset(dataset_name)
        return 200, {}, json.dumps(dataset.as_dict())

    @amzn_request_id
    def dataset_response(self, request: Any, full_url: str, headers: Any) -> Union[str, TYPE_RESPONSE]:  # type: ignore[return]
        self.setup_class(request, full_url, headers)

        dataset_name = self.parsed_url.path.split("/")[-1]

        if request.method == "POST":
            return self.create_dataset()
        elif request.method == "GET" and dataset_name:
            return self.describe_dataset(dataset_name)
        elif request.method == "GET":
            return self.list_datasets()
        elif request.method == "DELETE":
            return self.delete_dataset(dataset_name)
        elif request.method == "PUT":
            return self.update_dataset(dataset_name)

    # endregion

    # region Jobs
    @amzn_request_id
    def list_jobs(self, request: Any, full_url: str, headers: Any) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_ListJobs.html
        self.setup_class(request, full_url, headers)
        dataset_name = self._get_param("datasetName")
        project_name = self._get_param("projectName")
        next_token = self._get_param("NextToken", self._get_param("nextToken"))
        max_results = self._get_int_param(
            "MaxResults", self._get_int_param("maxResults")
        )

        # pylint: disable=unexpected-keyword-arg, unbalanced-tuple-unpacking
        job_list, next_token = self.databrew_backend.list_jobs(
            dataset_name=dataset_name,
            project_name=project_name,
            next_token=next_token,
            max_results=max_results,
        )
        return json.dumps(
            {
                "Jobs": [job.as_dict() for job in job_list],
                "NextToken": next_token,
            }
        )

    def get_job_response(self, job_name: str) -> TYPE_RESPONSE:
        job = self.databrew_backend.describe_job(job_name)
        return 200, {}, json.dumps(job.as_dict())

    def delete_job_response(self, job_name: str) -> TYPE_RESPONSE:
        self.databrew_backend.delete_job(job_name)
        return 200, {}, json.dumps({"Name": job_name})

    @amzn_request_id
    def job_response(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)

        job_name = self.parsed_url.path.rstrip("/").rsplit("/", 1)[1]

        if request.method == "GET":
            return self.get_job_response(job_name)
        elif request.method == "DELETE":
            return self.delete_job_response(job_name)

    @amzn_request_id
    def create_profile_job(self) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_CreateProfileJob.html
        kwargs = {
            "dataset_name": self._get_param("DatasetName"),
            "name": self._get_param("Name"),
            "output_location": self._get_param("OutputLocation"),
            "role_arn": self._get_param("RoleArn"),
            "configuration": self._get_param("Configuration"),
            "encryption_key_arn": self._get_param("EncryptionKeyArn"),
            "encryption_mode": self._get_param("EncryptionMode"),
            "job_sample": self._get_param("JobSample"),
            "log_subscription": self._get_param("LogSubscription"),
            "max_capacity": self._get_int_param("MaxCapacity"),
            "max_retries": self._get_int_param("MaxRetries"),
            "tags": self._get_param("Tags"),
            "timeout": self._get_int_param("Timeout"),
            "validation_configurations": self._get_param("ValidationConfigurations"),
        }
        return json.dumps(self.databrew_backend.create_profile_job(**kwargs).as_dict())

    def update_profile_job_response(self, name: str) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_UpdateProfileJob.html
        kwargs = {
            "name": name,
            "output_location": self._get_param("OutputLocation"),
            "role_arn": self._get_param("RoleArn"),
            "configuration": self._get_param("Configuration"),
            "encryption_key_arn": self._get_param("EncryptionKeyArn"),
            "encryption_mode": self._get_param("EncryptionMode"),
            "job_sample": self._get_param("JobSample"),
            "log_subscription": self._get_param("LogSubscription"),
            "max_capacity": self._get_int_param("MaxCapacity"),
            "max_retries": self._get_int_param("MaxRetries"),
            "timeout": self._get_int_param("Timeout"),
            "validation_configurations": self._get_param("ValidationConfigurations"),
        }
        return json.dumps(self.databrew_backend.update_profile_job(**kwargs).as_dict())

    @amzn_request_id
    def create_recipe_job(self) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_CreateRecipeJob.html
        kwargs = {
            "name": self._get_param("Name"),
            "role_arn": self._get_param("RoleArn"),
            "database_outputs": self._get_param("DatabaseOutputs"),
            "data_catalog_outputs": self._get_param("DataCatalogOutputs"),
            "dataset_name": self._get_param("DatasetName"),
            "encryption_key_arn": self._get_param("EncryptionKeyArn"),
            "encryption_mode": self._get_param("EncryptionMode"),
            "log_subscription": self._get_param("LogSubscription"),
            "max_capacity": self._get_int_param("MaxCapacity"),
            "max_retries": self._get_int_param("MaxRetries"),
            "outputs": self._get_param("Outputs"),
            "project_name": self._get_param("ProjectName"),
            "recipe_reference": self._get_param("RecipeReference"),
            "tags": self._get_param("Tags"),
            "timeout": self._get_int_param("Timeout"),
        }
        return json.dumps(self.databrew_backend.create_recipe_job(**kwargs).as_dict())

    @amzn_request_id
    def update_recipe_job_response(self, name: str) -> str:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_UpdateRecipeJob.html
        kwargs = {
            "name": name,
            "role_arn": self._get_param("RoleArn"),
            "database_outputs": self._get_param("DatabaseOutputs"),
            "data_catalog_outputs": self._get_param("DataCatalogOutputs"),
            "encryption_key_arn": self._get_param("EncryptionKeyArn"),
            "encryption_mode": self._get_param("EncryptionMode"),
            "log_subscription": self._get_param("LogSubscription"),
            "max_capacity": self._get_int_param("MaxCapacity"),
            "max_retries": self._get_int_param("MaxRetries"),
            "outputs": self._get_param("Outputs"),
            "timeout": self._get_int_param("Timeout"),
        }
        return json.dumps(self.databrew_backend.update_recipe_job(**kwargs).as_dict())

    @amzn_request_id
    def profile_job_response(self, request: Any, full_url: str, headers: Any) -> str:  # type: ignore[return]
        self.setup_class(request, full_url, headers)

        job_name = self.parsed_url.path.rstrip("/").rsplit("/", 1)[1]

        if request.method == "PUT":
            return self.update_profile_job_response(job_name)

    @amzn_request_id
    def recipe_job_response(self, request: Any, full_url: str, headers: Any) -> str:  # type: ignore[return]
        self.setup_class(request, full_url, headers)

        job_name = self.parsed_url.path.rstrip("/").rsplit("/", 1)[1]

        if request.method == "PUT":
            return self.update_recipe_job_response(job_name)

    # endregion



# type: ignore
import abc
from abc import abstractmethod
from collections import deque

from moto.dynamodb.models import DynamoType
from ..exceptions import DuplicateUpdateExpression, TooManyAddClauses


class Node(metaclass=abc.ABCMeta):
    def __init__(self, children=None):
        self.type = self.__class__.__name__
        assert children is None or isinstance(children, list)
        self.children = children
        self.parent = None

        if isinstance(children, list):
            for child in children:
                if isinstance(child, Node):
                    child.set_parent(self)

    def set_parent(self, parent_node):
        self.parent = parent_node

    def validate(self) -> None:
        if self.type == "UpdateExpression":
            nr_of_clauses = len(self.find_clauses([UpdateExpressionAddClause]))
            if nr_of_clauses > 1:
                raise TooManyAddClauses()
            set_actions = self.find_clauses([UpdateExpressionSetAction])
            # set_attributes = ["attr", "map.attr", attr.list[2], ..]
            set_attributes = [s.children[0].to_str() for s in set_actions]
            # We currently only check for duplicates
            # We should also check for partial duplicates, i.e. [attr, attr.sub] is also invalid
            if len(set_attributes) != len(set(set_attributes)):
                raise DuplicateUpdateExpression(set_attributes)

    def normalize(self):
        """
        Flatten the Add-/Delete-/Remove-/Set-Action children within this Node
        """
        if self.type == "UpdateExpression":
            # We can have multiple REMOVE attr[idx] expressions, such as attr[i] and attr[i+2]
            # If we remove attr[i] first, attr[i+2] suddenly refers to a different item
            # So we sort them in reverse order - we can remove attr[i+2] first, attr[i] still refers to the same item

            # Behaviour that is unknown, for now:
            # What happens if we SET and REMOVE on the same list - what takes precedence?
            # We're assuming this is executed in original order

            remove_actions = []
            sorted_actions = []
            possible_clauses = [
                UpdateExpressionAddAction,
                UpdateExpressionDeleteAction,
                UpdateExpressionRemoveAction,
                UpdateExpressionSetAction,
            ]
            for action in self.find_clauses(possible_clauses):
                if isinstance(action, UpdateExpressionRemoveAction):
                    # Keep these separate for now
                    remove_actions.append(action)
                else:
                    if len(remove_actions) > 0:
                        # Remove-actions were found earlier
                        # Now that we have other action-types, that means we've found all possible Remove-actions
                        # Sort them appropriately
                        sorted_actions.extend(sorted(remove_actions, reverse=True))
                        remove_actions.clear()
                    # Add other actions by insertion order
                    sorted_actions.append(action)
            # Remove actions were found last
            if len(remove_actions) > 0:
                sorted_actions.extend(sorted(remove_actions, reverse=True))

            self.children = sorted_actions

    def find_clauses(self, clause_types):
        clauses = []
        for child in self.children or []:
            if type(child) in clause_types:
                clauses.append(child)
            elif isinstance(child, Expression):
                clauses.extend(child.find_clauses(clause_types))
        return clauses


class LeafNode(Node):
    """A LeafNode is a Node where none of the children are Nodes themselves."""

    def __init__(self, children=None):
        super().__init__(children)


class Expression(Node, metaclass=abc.ABCMeta):
    """
    Abstract Syntax Tree representing the expression

    For the Grammar start here and jump down into the classes at the righ-hand side to look further. Nodes marked with
    a star are abstract and won't appear in the final AST.

    Expression* => UpdateExpression
    Expression* => ConditionExpression
    """


class UpdateExpression(Expression):
    """
    UpdateExpression => UpdateExpressionClause*
    UpdateExpression => UpdateExpressionClause* UpdateExpression
    """


class UpdateExpressionClause(UpdateExpression, metaclass=abc.ABCMeta):
    """
    UpdateExpressionClause* => UpdateExpressionSetClause
    UpdateExpressionClause* => UpdateExpressionRemoveClause
    UpdateExpressionClause* => UpdateExpressionAddClause
    UpdateExpressionClause* => UpdateExpressionDeleteClause
    """


class UpdateExpressionSetClause(UpdateExpressionClause):
    """
    UpdateExpressionSetClause => SET SetActions
    """


class UpdateExpressionSetActions(UpdateExpressionClause):
    """
    UpdateExpressionSetClause => SET SetActions

    SetActions => SetAction
    SetActions => SetAction , SetActions

    """


class UpdateExpressionSetAction(UpdateExpressionClause):
    """
    SetAction => Path = Value
    """


class UpdateExpressionRemoveActions(UpdateExpressionClause):
    """
    UpdateExpressionSetClause => REMOVE RemoveActions

    RemoveActions => RemoveAction
    RemoveActions => RemoveAction , RemoveActions
    """


class UpdateExpressionRemoveAction(UpdateExpressionClause):
    """
    RemoveAction => Path
    """

    def _get_value(self):
        expression_path = self.children[0]
        expression_selector = expression_path.children[-1]
        return expression_selector.children[0]

    def __lt__(self, other):
        self_value = self._get_value()

        return self_value < other._get_value()


class UpdateExpressionAddActions(UpdateExpressionClause):
    """
    UpdateExpressionAddClause => ADD RemoveActions

    AddActions => AddAction
    AddActions => AddAction , AddActions
    """


class UpdateExpressionAddAction(UpdateExpressionClause):
    """
    AddAction => Path Value
    """


class UpdateExpressionDeleteActions(UpdateExpressionClause):
    """
    UpdateExpressionDeleteClause => DELETE RemoveActions

    DeleteActions => DeleteAction
    DeleteActions => DeleteAction , DeleteActions
    """


class UpdateExpressionDeleteAction(UpdateExpressionClause):
    """
    DeleteAction => Path Value
    """


class UpdateExpressionPath(UpdateExpressionClause):
    def to_str(self):
        return "".join(x.to_str() for x in self.children)


class UpdateExpressionValue(UpdateExpressionClause):
    """
    Value => Operand
    Value => Operand + Value
    Value => Operand - Value
    """


class UpdateExpressionGroupedValue(UpdateExpressionClause):
    """
    GroupedValue => ( Value )
    """


class UpdateExpressionRemoveClause(UpdateExpressionClause):
    """
    UpdateExpressionRemoveClause => REMOVE RemoveActions
    """


class UpdateExpressionAddClause(UpdateExpressionClause):
    """
    UpdateExpressionAddClause => ADD AddActions
    """


class UpdateExpressionDeleteClause(UpdateExpressionClause):
    """
    UpdateExpressionDeleteClause => DELETE DeleteActions
    """


class ExpressionPathDescender(Node):
    """Node identifying descender into nested structure (.) in expression"""

    def to_str(self):
        return "."


class ExpressionSelector(LeafNode):
    """Node identifying selector [selection_index] in expresion"""

    def __init__(self, selection_index):
        try:
            super().__init__(children=[int(selection_index)])
        except ValueError:
            assert (
                False
            ), "Expression selector must be an int, this is a bug in the moto library."

    def get_index(self):
        return self.children[0]

    def to_str(self):
        return f"[{self.get_index()}]"


class ExpressionAttribute(LeafNode):
    """An attribute identifier as used in the DDB item"""

    def __init__(self, attribute):
        super().__init__(children=[attribute])

    def get_attribute_name(self):
        return self.children[0]

    def to_str(self):
        return self.get_attribute_name()


class ExpressionAttributeName(LeafNode):
    """An ExpressionAttributeName is an alias for an attribute identifier"""

    def __init__(self, attribute_name):
        super().__init__(children=[attribute_name])

    def get_attribute_name_placeholder(self):
        return self.children[0]

    def to_str(self):
        return self.get_attribute_name_placeholder()


class ExpressionAttributeValue(LeafNode):
    """An ExpressionAttributeValue is an alias for an value"""

    def __init__(self, value):
        super().__init__(children=[value])

    def get_value_name(self):
        return self.children[0]


class ExpressionValueOperator(LeafNode):
    """An ExpressionValueOperator is an operation that works on 2 values"""

    def __init__(self, value):
        super().__init__(children=[value])

    def get_operator(self):
        return self.children[0]


class UpdateExpressionFunction(Node):
    """
    A Node representing a function of an Update Expression. The first child is the function name the others are the
    arguments.
    """

    def get_function_name(self):
        return self.children[0]

    def get_nth_argument(self, n=1):
        """Return nth element where n is a 1-based index."""
        assert n >= 1
        return self.children[n]


class DDBTypedValue(Node):
    """
    A node representing a DDBTyped value. This can be any structure as supported by DyanmoDB. The node only has 1 child
    which is the value of type `DynamoType`.
    """

    def __init__(self, value):
        assert isinstance(value, DynamoType), "DDBTypedValue must be of DynamoType"
        super().__init__(children=[value])

    def get_value(self):
        return self.children[0]


class NoneExistingPath(LeafNode):
    """A placeholder for Paths that did not exist in the Item."""

    def __init__(self, creatable=False):
        super().__init__(children=[creatable])

    def is_creatable(self):
        """Can this path be created if need be. For example path creating element in a dictionary or creating a new
        attribute under root level of an item."""
        return self.children[0]


class DepthFirstTraverser(object):
    """
    Helper class that allows depth first traversal and to implement custom processing for certain AST nodes. The
    processor of a node must return the new resulting node. This node will be placed in the tree. Processing of a
    node using this traverser should therefore only transform child nodes.  The returned node will get the same parent
    as the node before processing had.
    """

    @abstractmethod
    def _processing_map(self):
        """
        A map providing a processing function per node class type to a function that takes in a Node object and
        processes it. A Node can only be processed by a single function and they are considered in order. Therefore if
        multiple classes from a single class hierarchy strain are used the more specific classes have to be put before
        the less specific ones. That requires overriding `nodes_to_be_processed`. If no multiple classes form a single
        class hierarchy strain are used the default implementation of `nodes_to_be_processed` should be OK.
        Returns:
            dict: Mapping a Node Class to a processing function.
        """
        pass

    def nodes_to_be_processed(self):
        """Cached accessor for getting Node types that need to be processed."""
        return tuple(k for k in self._processing_map().keys())

    def process(self, node):
        """Process a Node"""
        for class_key, processor in self._processing_map().items():
            if isinstance(node, class_key):
                return processor(node)

    def pre_processing_of_child(self, parent_node, child_id):
        """Hook that is called pre-processing of the child at position `child_id`"""
        pass

    def traverse_node_recursively(self, node, child_id=-1):
        """
        Traverse nodes depth first processing nodes bottom up (if root node is considered the top).

        Args:
            node(Node): The node which is the last node to be processed but which allows to identify all the
                             work (which is in the children)
            child_id(int): The index in the list of children from the parent that this node corresponds to

        Returns:
            Node: The node of the new processed AST
        """
        if isinstance(node, Node):
            parent_node = node.parent
            if node.children is not None:
                for i, child_node in enumerate(node.children):
                    self.pre_processing_of_child(node, i)
                    self.traverse_node_recursively(child_node, i)
            # noinspection PyTypeChecker
            if isinstance(node, self.nodes_to_be_processed()):
                node = self.process(node)
                node.parent = parent_node
                parent_node.children[child_id] = node
        return node

    def traverse(self, node):
        return self.traverse_node_recursively(node)


class NodeDepthLeftTypeFetcher(object):
    """Helper class to fetch a node of a specific type. Depth left-first traversal"""

    def __init__(self, node_type, root_node):
        assert issubclass(node_type, Node)
        self.node_type = node_type
        self.root_node = root_node
        self.queue = deque()
        self.add_nodes_left_to_right_depth_first(self.root_node)

    def add_nodes_left_to_right_depth_first(self, node):
        if isinstance(node, Node) and node.children is not None:
            for child_node in node.children:
                self.add_nodes_left_to_right_depth_first(child_node)
                self.queue.append(child_node)
        self.queue.append(node)

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        while len(self.queue) > 0:
            candidate = self.queue.popleft()
            if isinstance(candidate, self.node_type):
                return candidate
        raise StopIteration



import string
from typing import Optional

from email.utils import parseaddr
from moto.moto_api._internal import mock_random as random


def random_hex(length: int) -> str:
    return "".join(random.choice(string.ascii_lowercase) for x in range(length))


def get_random_message_id() -> str:
    return f"{random_hex(16)}-{random_hex(8)}-{random_hex(4)}-{random_hex(4)}-{random_hex(4)}-{random_hex(12)}-{random_hex(6)}"


def is_valid_address(addr: str) -> Optional[str]:
    _, address = parseaddr(addr)
    address_parts = address.split("@")
    if len(address_parts) != 2 or not address_parts[1]:
        return "Missing domain"
    return None



from moto.core.responses import BaseResponse
from .models import iotdata_backends, IoTDataPlaneBackend
import json
from typing import Any
from urllib.parse import unquote


class IoTDataPlaneResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="iot-data")

    def setup_class(
        self, request: Any, full_url: str, headers: Any, use_raw_body: bool = False
    ) -> None:
        super().setup_class(request, full_url, headers, use_raw_body=True)

    def _get_action(self) -> str:
        if self.path and self.path.startswith("/topics/"):
            # Special usecase - there is no way identify this action, besides the URL
            return "publish"
        return super()._get_action()

    @property
    def iotdata_backend(self) -> IoTDataPlaneBackend:
        return iotdata_backends[self.current_account][self.region]

    def update_thing_shadow(self) -> str:
        thing_name = self._get_param("thingName")
        shadow_name = self.querystring.get("name", [None])[0]
        payload = self.iotdata_backend.update_thing_shadow(
            thing_name=thing_name,
            payload=self.body,
            shadow_name=shadow_name,
        )
        return json.dumps(payload.to_response_dict())

    def get_thing_shadow(self) -> str:
        thing_name = self._get_param("thingName")
        shadow_name = self.querystring.get("name", [None])[0]
        payload = self.iotdata_backend.get_thing_shadow(
            thing_name=thing_name, shadow_name=shadow_name
        )
        return json.dumps(payload.to_dict())

    def delete_thing_shadow(self) -> str:
        thing_name = self._get_param("thingName")
        shadow_name = self.querystring.get("name", [None])[0]
        payload = self.iotdata_backend.delete_thing_shadow(
            thing_name=thing_name, shadow_name=shadow_name
        )
        return json.dumps(payload.to_dict())

    def publish(self) -> str:
        topic = self.path.split("/topics/")[-1]
        # a uri parameter containing forward slashes is not correctly url encoded when we're running in server mode.
        # https://github.com/pallets/flask/issues/900
        topic = unquote(topic) if "%" in topic else topic
        self.iotdata_backend.publish(topic=topic, payload=self.body)
        return json.dumps(dict())

    def list_named_shadows_for_thing(self) -> str:
        thing_name = self._get_param("thingName")
        shadows = self.iotdata_backend.list_named_shadows_for_thing(thing_name)
        return json.dumps({"results": [shadow.to_dict() for shadow in shadows]})



from typing import Any
from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse
from .models import guardduty_backends, GuardDutyBackend
import json
from urllib.parse import unquote


class GuardDutyResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="guardduty")

    @property
    def guardduty_backend(self) -> GuardDutyBackend:
        return guardduty_backends[self.current_account][self.region]

    def filter(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "GET":
            return self.get_filter()
        elif request.method == "DELETE":
            return self.delete_filter()
        elif request.method == "POST":
            return self.update_filter()

    def filters(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "POST":
            return self.create_filter()

    def detectors(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:
        self.setup_class(request, full_url, headers)
        if request.method == "POST":
            return self.create_detector()
        elif request.method == "GET":
            return self.list_detectors()
        else:
            return 404, {}, ""

    def detector(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "GET":
            return self.get_detector()
        elif request.method == "DELETE":
            return self.delete_detector()
        elif request.method == "POST":
            return self.update_detector()

    def create_filter(self) -> TYPE_RESPONSE:
        detector_id = self.path.split("/")[-2]
        name = self._get_param("name")
        action = self._get_param("action")
        description = self._get_param("description")
        finding_criteria = self._get_param("findingCriteria")
        rank = self._get_param("rank")

        self.guardduty_backend.create_filter(
            detector_id, name, action, description, finding_criteria, rank
        )
        return 200, {}, json.dumps({"name": name})

    def create_detector(self) -> TYPE_RESPONSE:
        enable = self._get_param("enable")
        finding_publishing_frequency = self._get_param("findingPublishingFrequency")
        data_sources = self._get_param("dataSources")
        tags = self._get_param("tags")

        detector_id = self.guardduty_backend.create_detector(
            enable, finding_publishing_frequency, data_sources, tags
        )

        return 200, {}, json.dumps(dict(detectorId=detector_id))

    def delete_detector(self) -> TYPE_RESPONSE:
        detector_id = self.path.split("/")[-1]

        self.guardduty_backend.delete_detector(detector_id)
        return 200, {}, "{}"

    def delete_filter(self) -> TYPE_RESPONSE:
        detector_id = self.path.split("/")[-3]
        filter_name = unquote(self.path.split("/")[-1])

        self.guardduty_backend.delete_filter(detector_id, filter_name)
        return 200, {}, "{}"

    def enable_organization_admin_account(
        self, request: Any, full_url: str, headers: Any
    ) -> TYPE_RESPONSE:
        self.setup_class(request, full_url, headers)

        admin_account = self._get_param("adminAccountId")
        self.guardduty_backend.enable_organization_admin_account(admin_account)

        return 200, {}, "{}"

    def list_organization_admin_accounts(
        self, request: Any, full_url: str, headers: Any
    ) -> TYPE_RESPONSE:
        self.setup_class(request, full_url, headers)

        account_ids = self.guardduty_backend.list_organization_admin_accounts()

        return (
            200,
            {},
            json.dumps(
                {
                    "adminAccounts": [
                        {"adminAccountId": account_id, "adminStatus": "ENABLED"}
                        for account_id in account_ids
                    ]
                }
            ),
        )

    def list_detectors(self) -> TYPE_RESPONSE:
        detector_ids = self.guardduty_backend.list_detectors()

        return 200, {}, json.dumps({"detectorIds": detector_ids})

    def get_detector(self) -> TYPE_RESPONSE:
        detector_id = self.path.split("/")[-1]

        detector = self.guardduty_backend.get_detector(detector_id)
        return 200, {}, json.dumps(detector.to_json())

    def get_filter(self) -> TYPE_RESPONSE:
        detector_id = self.path.split("/")[-3]
        filter_name = unquote(self.path.split("/")[-1])

        _filter = self.guardduty_backend.get_filter(detector_id, filter_name)
        return 200, {}, json.dumps(_filter.to_json())

    def update_detector(self) -> TYPE_RESPONSE:
        detector_id = self.path.split("/")[-1]
        enable = self._get_param("enable")
        finding_publishing_frequency = self._get_param("findingPublishingFrequency")
        data_sources = self._get_param("dataSources")

        self.guardduty_backend.update_detector(
            detector_id, enable, finding_publishing_frequency, data_sources
        )
        return 200, {}, "{}"

    def update_filter(self) -> TYPE_RESPONSE:
        detector_id = self.path.split("/")[-3]
        filter_name = unquote(self.path.split("/")[-1])
        action = self._get_param("action")
        description = self._get_param("description")
        finding_criteria = self._get_param("findingCriteria")
        rank = self._get_param("rank")

        self.guardduty_backend.update_filter(
            detector_id,
            filter_name,
            action=action,
            description=description,
            finding_criteria=finding_criteria,
            rank=rank,
        )
        return 200, {}, json.dumps({"name": filter_name})



from collections import defaultdict
from datetime import datetime
import re
import string
import time
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

from openapi_spec_validator import validate_spec
import requests
import responses

try:
    from openapi_spec_validator.validation.exceptions import OpenAPIValidationError
except ImportError:
    # OpenAPI Spec Validator < 0.5.0
    from openapi_spec_validator.exceptions import OpenAPIValidationError  # type: ignore
from moto.core import BaseBackend, BackendDict, BaseModel, CloudFormationModel
from .utils import create_id, to_path
from moto.core.utils import path_url
from .exceptions import (
    BadRequestException,
    ConflictException,
    DeploymentNotFoundException,
    ApiKeyNotFoundException,
    UsagePlanNotFoundException,
    AwsProxyNotAllowed,
    CrossAccountNotAllowed,
    IntegrationMethodNotDefined,
    InvalidArn,
    InvalidIntegrationArn,
    InvalidHttpEndpoint,
    InvalidOpenAPIDocumentException,
    InvalidOpenApiDocVersionException,
    InvalidOpenApiModeException,
    InvalidResourcePathException,
    AuthorizerNotFoundException,
    StageNotFoundException,
    ResourceIdNotFoundException,
    RoleNotSpecified,
    NoIntegrationDefined,
    NoIntegrationResponseDefined,
    NoMethodDefined,
    ApiKeyAlreadyExists,
    DomainNameNotFound,
    InvalidDomainName,
    InvalidRestApiId,
    InvalidModelName,
    RestAPINotFound,
    RequestValidatorNotFound,
    ModelNotFound,
    ApiKeyValueMinLength,
    InvalidBasePathException,
    InvalidRestApiIdForBasePathMappingException,
    InvalidStageException,
    BasePathConflictException,
    BasePathNotFoundException,
    StageStillActive,
    VpcLinkNotFound,
    ValidationException,
    GatewayResponseNotFound,
)
from ..core.models import responses_mock
from moto.apigateway.exceptions import MethodNotFoundException
from moto.moto_api._internal import mock_random as random

STAGE_URL = "https://{api_id}.execute-api.{region_name}.amazonaws.com/{stage_name}"


class Deployment(CloudFormationModel):
    def __init__(self, deployment_id: str, name: str, description: str = ""):
        self.id = deployment_id
        self.stage_name = name
        self.description = description
        self.created_date = int(time.time())

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "stageName": self.stage_name,
            "description": self.description,
            "createdDate": self.created_date,
        }

    @staticmethod
    def cloudformation_name_type() -> str:
        return "Deployment"

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::ApiGateway::Deployment"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Dict[str, Any],
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "Deployment":
        properties = cloudformation_json["Properties"]
        rest_api_id = properties["RestApiId"]
        name = properties.get("StageName")
        desc = properties.get("Description", "")
        backend: "APIGatewayBackend" = apigateway_backends[account_id][region_name]
        return backend.create_deployment(
            function_id=rest_api_id, name=name, description=desc
        )


class IntegrationResponse(BaseModel):
    def __init__(
        self,
        status_code: Union[str, int],
        selection_pattern: Optional[str] = None,
        response_templates: Optional[Dict[str, Any]] = None,
        response_parameters: Optional[Dict[str, str]] = None,
        content_handling: Optional[Any] = None,
    ):
        if response_templates is None:
            # response_templates = {"application/json": None}  # Note: removed for compatibility with TF
            response_templates = {}
        for key in response_templates.keys():
            response_templates[key] = (
                response_templates[key] or None
            )  # required for compatibility with TF
        self.response_templates = response_templates
        self.status_code = status_code
        self.selection_pattern = selection_pattern
        self.response_parameters = response_parameters
        self.content_handling = content_handling

    def to_json(self) -> Dict[str, Any]:
        resp = {
            "responseTemplates": self.response_templates,
            "statusCode": self.status_code,
        }
        if self.selection_pattern:
            resp["selectionPattern"] = self.selection_pattern
        if self.content_handling:
            resp["contentHandling"] = self.content_handling
        if self.response_parameters:
            resp["responseParameters"] = self.response_parameters
        return resp


class Integration(BaseModel):
    def __init__(
        self,
        integration_type: str,
        uri: str,
        http_method: str,
        request_templates: Optional[Dict[str, Any]] = None,
        passthrough_behavior: Optional[str] = "WHEN_NO_MATCH",
        cache_key_parameters: Optional[List[str]] = None,
        tls_config: Optional[Dict[str, Any]] = None,
        cache_namespace: Optional[str] = None,
        timeout_in_millis: Optional[str] = None,
        request_parameters: Optional[Dict[str, Any]] = None,
        content_handling: Optional[str] = None,
        credentials: Optional[str] = None,
        connection_type: Optional[str] = None,
    ):
        self.integration_type = integration_type
        self.uri = uri
        self.http_method = http_method if integration_type != "MOCK" else None
        self.passthrough_behaviour = passthrough_behavior
        self.cache_key_parameters: List[str] = cache_key_parameters or []
        self.request_templates = request_templates
        self.tls_config = tls_config
        self.cache_namespace = cache_namespace
        self.timeout_in_millis = timeout_in_millis
        self.request_parameters = request_parameters
        self.content_handling = content_handling
        self.credentials = credentials
        self.connection_type = connection_type
        self.integration_responses: Optional[Dict[str, IntegrationResponse]] = None

    def to_json(self) -> Dict[str, Any]:
        int_responses: Optional[Dict[str, Any]] = None
        if self.integration_responses is not None:
            int_responses = {
                k: v.to_json() for k, v in self.integration_responses.items()
            }
        return {
            "type": self.integration_type,
            "uri": self.uri,
            "httpMethod": self.http_method,
            "passthroughBehavior": self.passthrough_behaviour,
            "cacheKeyParameters": self.cache_key_parameters,
            "requestTemplates": self.request_templates,
            "integrationResponses": int_responses,
            "tlsConfig": self.tls_config,
            "cacheNamespace": self.cache_namespace,
            "timeoutInMillis": self.timeout_in_millis,
            "requestParameters": self.request_parameters,
            "contentHandling": self.content_handling,
            "credentials": self.credentials,
            "connectionType": self.connection_type,
        }

    def create_integration_response(
        self,
        status_code: str,
        selection_pattern: str,
        response_templates: Dict[str, str],
        response_parameters: Dict[str, str],
        content_handling: str,
    ) -> IntegrationResponse:
        integration_response = IntegrationResponse(
            status_code,
            selection_pattern,
            response_templates or None,
            response_parameters,
            content_handling,
        )
        if self.integration_responses is None:
            self.integration_responses = {}
        self.integration_responses[status_code] = integration_response
        return integration_response

    def get_integration_response(self, status_code: str) -> IntegrationResponse:
        result = (self.integration_responses or {}).get(status_code)
        if not result:
            raise NoIntegrationResponseDefined()
        return result

    def delete_integration_response(self, status_code: str) -> IntegrationResponse:
        return (self.integration_responses or {}).pop(status_code, None)  # type: ignore[arg-type]


class MethodResponse(BaseModel):
    def __init__(
        self,
        status_code: str,
        response_models: Dict[str, str],
        response_parameters: Dict[str, Dict[str, str]],
    ):
        self.status_code = status_code
        self.response_models = response_models
        self.response_parameters = response_parameters

    def to_json(self) -> Dict[str, Any]:
        return {
            "statusCode": self.status_code,
            "responseModels": self.response_models,
            "responseParameters": self.response_parameters,
        }


class Method(CloudFormationModel):
    def __init__(
        self, method_type: str, authorization_type: Optional[str], **kwargs: Any
    ):
        self.http_method = method_type
        self.authorization_type = authorization_type
        self.authorizer_id = kwargs.get("authorizer_id")
        self.authorization_scopes = kwargs.get("authorization_scopes")
        self.api_key_required = kwargs.get("api_key_required") or False
        self.request_parameters = kwargs.get("request_parameters")
        self.request_models = kwargs.get("request_models")
        self.method_integration: Optional[Integration] = None
        self.operation_name = kwargs.get("operation_name")
        self.request_validator_id = kwargs.get("request_validator_id")
        self.method_responses: Dict[str, MethodResponse] = {}

    def to_json(self) -> Dict[str, Any]:
        return {
            "httpMethod": self.http_method,
            "authorizationType": self.authorization_type,
            "authorizerId": self.authorizer_id,
            "authorizationScopes": self.authorization_scopes,
            "apiKeyRequired": self.api_key_required,
            "requestParameters": self.request_parameters,
            "requestModels": self.request_models,
            "methodIntegration": self.method_integration.to_json()
            if self.method_integration
            else None,
            "operationName": self.operation_name,
            "requestValidatorId": self.request_validator_id,
            "methodResponses": {
                k: v.to_json() for k, v in self.method_responses.items()
            },
        }

    @staticmethod
    def cloudformation_name_type() -> str:
        return "Method"

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::ApiGateway::Method"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Dict[str, Any],
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "Method":
        properties = cloudformation_json["Properties"]
        rest_api_id = properties["RestApiId"]
        resource_id = properties["ResourceId"]
        method_type = properties["HttpMethod"]
        auth_type = properties["AuthorizationType"]
        key_req = properties.get("ApiKeyRequired")
        backend = apigateway_backends[account_id][region_name]
        m = backend.put_method(
            function_id=rest_api_id,
            resource_id=resource_id,
            method_type=method_type,
            authorization_type=auth_type,
            api_key_required=key_req,
        )
        int_method = properties["Integration"]["IntegrationHttpMethod"]
        int_type = properties["Integration"]["Type"]
        int_uri = properties["Integration"]["Uri"]
        backend.put_integration(
            function_id=rest_api_id,
            resource_id=resource_id,
            method_type=method_type,
            integration_type=int_type,
            uri=int_uri,
            integration_method=int_method,
        )
        return m

    def create_response(
        self,
        response_code: str,
        response_models: Dict[str, str],
        response_parameters: Dict[str, Dict[str, str]],
    ) -> MethodResponse:
        method_response = MethodResponse(
            response_code, response_models, response_parameters
        )
        self.method_responses[response_code] = method_response
        return method_response

    def get_response(self, response_code: str) -> Optional[MethodResponse]:
        return self.method_responses.get(response_code)

    def delete_response(self, response_code: str) -> Optional[MethodResponse]:
        return self.method_responses.pop(response_code, None)


class Resource(CloudFormationModel):
    def __init__(
        self,
        resource_id: str,
        account_id: str,
        region_name: str,
        api_id: str,
        path_part: str,
        parent_id: Optional[str],
    ):
        self.id = resource_id
        self.account_id = account_id
        self.region_name = region_name
        self.api_id = api_id
        self.path_part = path_part
        self.parent_id = parent_id
        self.resource_methods: Dict[str, Method] = {}
        from .integration_parsers import IntegrationParser
        from .integration_parsers.aws_parser import TypeAwsParser
        from .integration_parsers.http_parser import TypeHttpParser
        from .integration_parsers.unknown_parser import TypeUnknownParser

        self.integration_parsers: Dict[str, IntegrationParser] = defaultdict(
            TypeUnknownParser
        )
        self.integration_parsers["HTTP"] = TypeHttpParser()
        self.integration_parsers["AWS"] = TypeAwsParser()

    def to_dict(self) -> Dict[str, Any]:
        response: Dict[str, Any] = {
            "path": self.get_path(),
            "id": self.id,
        }
        if self.resource_methods:
            response["resourceMethods"] = {
                k: v.to_json() for k, v in self.resource_methods.items()
            }
        if self.parent_id:
            response["parentId"] = self.parent_id
            response["pathPart"] = self.path_part
        return response

    @property
    def physical_resource_id(self) -> str:
        return self.id

    @staticmethod
    def cloudformation_name_type() -> str:
        return "Resource"

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::ApiGateway::Resource"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Dict[str, Any],
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "Resource":
        properties = cloudformation_json["Properties"]
        api_id = properties["RestApiId"]
        parent = properties["ParentId"]
        path = properties["PathPart"]

        backend = apigateway_backends[account_id][region_name]
        if parent == api_id:
            # A Root path (/) is automatically created. Any new paths should use this as their parent
            resources = backend.get_resources(function_id=api_id)
            root_id = [resource for resource in resources if resource.path_part == "/"][
                0
            ].id
            parent = root_id
        return backend.create_resource(
            function_id=api_id, parent_resource_id=parent, path_part=path
        )

    def get_path(self) -> str:
        return self.get_parent_path() + self.path_part

    def get_parent_path(self) -> str:
        if self.parent_id:
            backend = apigateway_backends[self.account_id][self.region_name]
            parent = backend.get_resource(self.api_id, self.parent_id)
            parent_path = parent.get_path()
            if parent_path != "/":  # Root parent
                parent_path += "/"
            return parent_path
        else:
            return ""

    def get_response(
        self, request: requests.PreparedRequest
    ) -> Tuple[int, Union[str, bytes]]:
        integration = self.get_integration(str(request.method))
        integration_type = integration.integration_type  # type: ignore[union-attr]

        status, result = self.integration_parsers[integration_type].invoke(
            request, integration  # type: ignore[arg-type]
        )

        return status, result

    def add_method(
        self,
        method_type: str,
        authorization_type: Optional[str],
        api_key_required: Optional[bool],
        request_parameters: Any = None,
        request_models: Any = None,
        operation_name: Optional[str] = None,
        authorizer_id: Optional[str] = None,
        authorization_scopes: Any = None,
        request_validator_id: Any = None,
    ) -> Method:
        if authorization_scopes and not isinstance(authorization_scopes, list):
            authorization_scopes = [authorization_scopes]
        method = Method(
            method_type=method_type,
            authorization_type=authorization_type,
            api_key_required=api_key_required,
            request_parameters=request_parameters,
            request_models=request_models,
            operation_name=operation_name,
            authorizer_id=authorizer_id,
            authorization_scopes=authorization_scopes,
            request_validator_id=request_validator_id,
        )
        self.resource_methods[method_type] = method
        return method

    def get_method(self, method_type: str) -> Method:
        method = self.resource_methods.get(method_type)
        if not method:
            raise MethodNotFoundException()
        return method

    def delete_method(self, method_type: str) -> None:
        self.resource_methods.pop(method_type, None)

    def add_integration(
        self,
        method_type: str,
        integration_type: str,
        uri: str,
        request_templates: Optional[Dict[str, Any]] = None,
        passthrough_behavior: Optional[str] = None,
        integration_method: Optional[str] = None,
        tls_config: Optional[Dict[str, Any]] = None,
        cache_namespace: Optional[str] = None,
        timeout_in_millis: Optional[str] = None,
        request_parameters: Optional[Dict[str, Any]] = None,
        content_handling: Optional[str] = None,
        credentials: Optional[str] = None,
        connection_type: Optional[str] = None,
    ) -> Integration:
        integration_method = integration_method or method_type
        integration = Integration(
            integration_type,
            uri,
            integration_method,
            request_templates=request_templates,
            passthrough_behavior=passthrough_behavior,
            tls_config=tls_config,
            cache_namespace=cache_namespace,
            timeout_in_millis=timeout_in_millis,
            request_parameters=request_parameters,
            content_handling=content_handling,
            credentials=credentials,
            connection_type=connection_type,
        )
        self.resource_methods[method_type].method_integration = integration
        return integration

    def get_integration(self, method_type: str) -> Optional[Integration]:
        method = self.resource_methods.get(method_type)
        return method.method_integration if method else None

    def delete_integration(self, method_type: str) -> Integration:
        integration = self.resource_methods[method_type].method_integration
        self.resource_methods[method_type].method_integration = None
        return integration  # type: ignore[return-value]


class Authorizer(BaseModel):
    def __init__(
        self,
        authorizer_id: Optional[str],
        name: Optional[str],
        authorizer_type: Optional[str],
        **kwargs: Any,
    ):
        self.id = authorizer_id
        self.name = name
        self.type = authorizer_type
        self.provider_arns = kwargs.get("provider_arns")
        self.auth_type = kwargs.get("auth_type")
        self.authorizer_uri = kwargs.get("authorizer_uri")
        self.authorizer_credentials = kwargs.get("authorizer_credentials")
        self.identity_source = kwargs.get("identity_source")
        self.identity_validation_expression = kwargs.get(
            "identity_validation_expression"
        )
        self.authorizer_result_ttl = kwargs.get("authorizer_result_ttl")

    def to_json(self) -> Dict[str, Any]:
        dct = {
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "authorizerResultTtlInSeconds": self.authorizer_result_ttl,
        }
        if self.provider_arns:
            dct["providerARNs"] = self.provider_arns
        if self.auth_type:
            dct["authType"] = self.auth_type
        if self.authorizer_uri:
            dct["authorizerUri"] = self.authorizer_uri
        if self.authorizer_credentials:
            dct["authorizerCredentials"] = self.authorizer_credentials
        if self.identity_source:
            dct["identitySource"] = self.identity_source
        if self.identity_validation_expression:
            dct["identityValidationExpression"] = self.identity_validation_expression
        return dct

    def apply_operations(self, patch_operations: List[Dict[str, Any]]) -> "Authorizer":
        for op in patch_operations:
            if "/authorizerUri" in op["path"]:
                self.authorizer_uri = op["value"]
            elif "/authorizerCredentials" in op["path"]:
                self.authorizer_credentials = op["value"]
            elif "/authorizerResultTtlInSeconds" in op["path"]:
                self.authorizer_result_ttl = int(op["value"])
            elif "/authType" in op["path"]:
                self.auth_type = op["value"]
            elif "/identitySource" in op["path"]:
                self.identity_source = op["value"]
            elif "/identityValidationExpression" in op["path"]:
                self.identity_validation_expression = op["value"]
            elif "/name" in op["path"]:
                self.name = op["value"]
            elif "/providerARNs" in op["path"]:
                # TODO: add and remove
                raise Exception(f'Patch operation for "{op["path"]}" not implemented')
            elif "/type" in op["path"]:
                self.type = op["value"]
            else:
                raise Exception(f'Patch operation "{op["op"]}" not implemented')
        return self


class Stage(BaseModel):
    def __init__(
        self,
        name: Optional[str] = None,
        deployment_id: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
        description: str = "",
        cacheClusterEnabled: Optional[bool] = False,
        cacheClusterSize: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        tracing_enabled: Optional[bool] = None,
    ):
        self.name = name
        self.deployment_id = deployment_id
        self.method_settings: Dict[str, Any] = {}
        self.variables = variables or {}
        self.description = description
        self.cache_cluster_enabled = cacheClusterEnabled
        self.cache_cluster_status = "AVAILABLE" if cacheClusterEnabled else None
        self.cache_cluster_size = (
            str(cacheClusterSize) if cacheClusterSize is not None else None
        )
        self.tags = tags
        self.tracing_enabled = tracing_enabled
        self.access_log_settings: Optional[Dict[str, Any]] = None
        self.web_acl_arn: Optional[str] = None

    def to_json(self) -> Dict[str, Any]:
        dct: Dict[str, Any] = {
            "stageName": self.name,
            "deploymentId": self.deployment_id,
            "methodSettings": self.method_settings,
            "variables": self.variables,
            "description": self.description,
            "cacheClusterEnabled": self.cache_cluster_enabled,
            "accessLogSettings": self.access_log_settings,
        }
        if self.cache_cluster_status is not None:
            dct["cacheClusterStatus"] = self.cache_cluster_status
        if self.cache_cluster_enabled:
            if self.cache_cluster_size is not None:
                dct["cacheClusterSize"] = self.cache_cluster_size
            else:
                dct["cacheClusterSize"] = "0.5"
        if self.tags:
            dct["tags"] = self.tags
        if self.tracing_enabled is not None:
            dct["tracingEnabled"] = self.tracing_enabled
        if self.web_acl_arn is not None:
            dct["webAclArn"] = self.web_acl_arn
        return dct

    def apply_operations(self, patch_operations: List[Dict[str, Any]]) -> "Stage":
        for op in patch_operations:
            if "variables/" in op["path"]:
                self._apply_operation_to_variables(op)
            elif "/cacheClusterEnabled" in op["path"]:
                self.cache_cluster_enabled = self._str2bool(op["value"])
                if self.cache_cluster_enabled:
                    self.cache_cluster_status = "AVAILABLE"
                else:
                    self.cache_cluster_status = "NOT_AVAILABLE"
            elif "/cacheClusterSize" in op["path"]:
                self.cache_cluster_size = str(op["value"])
            elif "/description" in op["path"]:
                self.description = op["value"]
            elif "/deploymentId" in op["path"]:
                self.deployment_id = op["value"]
            elif op["op"] == "replace":
                if op["path"] == "/tracingEnabled":
                    self.tracing_enabled = self._str2bool(op["value"])
                elif op["path"].startswith("/accessLogSettings/"):
                    self.access_log_settings = self.access_log_settings or {}
                    self.access_log_settings[op["path"].split("/")[-1]] = op["value"]
                else:
                    # (e.g., path could be '/*/*/logging/loglevel')
                    split_path = op["path"].split("/", 3)
                    if len(split_path) != 4:
                        continue
                    self._patch_method_setting(
                        "/".join(split_path[1:3]), split_path[3], op["value"]
                    )
            elif op["op"] == "remove":
                if op["path"] == "/accessLogSettings":
                    self.access_log_settings = None
            else:
                raise ValidationException(
                    "Member must satisfy enum value set: [add, remove, move, test, replace, copy]"
                )
        return self

    def _patch_method_setting(
        self, resource_path_and_method: str, key: str, value: str
    ) -> None:
        updated_key = self._method_settings_translations(key)
        if updated_key is not None:
            if resource_path_and_method not in self.method_settings:
                self.method_settings[
                    resource_path_and_method
                ] = self._get_default_method_settings()
            self.method_settings[resource_path_and_method][
                updated_key
            ] = self._convert_to_type(updated_key, value)

    def _get_default_method_settings(self) -> Dict[str, Any]:
        return {
            "throttlingRateLimit": 1000.0,
            "dataTraceEnabled": False,
            "metricsEnabled": False,
            "unauthorizedCacheControlHeaderStrategy": "SUCCEED_WITH_RESPONSE_HEADER",
            "cacheTtlInSeconds": 300,
            "cacheDataEncrypted": True,
            "cachingEnabled": False,
            "throttlingBurstLimit": 2000,
            "requireAuthorizationForCacheControl": True,
        }

    def _method_settings_translations(self, key: str) -> Optional[str]:
        mappings = {
            "metrics/enabled": "metricsEnabled",
            "logging/loglevel": "loggingLevel",
            "logging/dataTrace": "dataTraceEnabled",
            "throttling/burstLimit": "throttlingBurstLimit",
            "throttling/rateLimit": "throttlingRateLimit",
            "caching/enabled": "cachingEnabled",
            "caching/ttlInSeconds": "cacheTtlInSeconds",
            "caching/dataEncrypted": "cacheDataEncrypted",
            "caching/requireAuthorizationForCacheControl": "requireAuthorizationForCacheControl",
            "caching/unauthorizedCacheControlHeaderStrategy": "unauthorizedCacheControlHeaderStrategy",
        }

        return mappings.get(key)

    def _str2bool(self, v: str) -> bool:
        return v.lower() == "true"

    def _convert_to_type(self, key: str, val: str) -> Union[str, int, float]:
        type_mappings = {
            "metricsEnabled": "bool",
            "loggingLevel": "str",
            "dataTraceEnabled": "bool",
            "throttlingBurstLimit": "int",
            "throttlingRateLimit": "float",
            "cachingEnabled": "bool",
            "cacheTtlInSeconds": "int",
            "cacheDataEncrypted": "bool",
            "requireAuthorizationForCacheControl": "bool",
            "unauthorizedCacheControlHeaderStrategy": "str",
        }

        if key in type_mappings:
            type_value = type_mappings[key]

            if type_value == "bool":
                return self._str2bool(val)
            elif type_value == "int":
                return int(val)
            elif type_value == "float":
                return float(val)
            else:
                return str(val)
        else:
            return str(val)

    def _apply_operation_to_variables(self, op: Dict[str, Any]) -> None:
        key = op["path"][op["path"].rindex("variables/") + 10 :]
        if op["op"] == "remove":
            self.variables.pop(key, None)
        elif op["op"] == "replace":
            self.variables[key] = op["value"]
        else:
            raise Exception(f'Patch operation "{op["op"]}" not implemented')


class ApiKey(BaseModel):
    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        enabled: bool = False,
        generateDistinctId: bool = False,  # pylint: disable=unused-argument
        value: Optional[str] = None,
        stageKeys: Optional[Any] = None,
        tags: Optional[List[Dict[str, str]]] = None,
        customerId: Optional[str] = None,
    ):
        self.id = create_id()
        self.value = value or "".join(
            random.sample(string.ascii_letters + string.digits, 40)
        )
        self.name = name
        self.customer_id = customerId
        self.description = description
        self.enabled = enabled
        self.created_date = self.last_updated_date = int(time.time())
        self.stage_keys = stageKeys or []
        self.tags = tags

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "value": self.value,
            "name": self.name,
            "customerId": self.customer_id,
            "description": self.description,
            "enabled": self.enabled,
            "createdDate": self.created_date,
            "lastUpdatedDate": self.last_updated_date,
            "stageKeys": self.stage_keys,
            "tags": self.tags,
        }

    def update_operations(self, patch_operations: List[Dict[str, Any]]) -> "ApiKey":
        for op in patch_operations:
            if op["op"] == "replace":
                if "/name" in op["path"]:
                    self.name = op["value"]
                elif "/customerId" in op["path"]:
                    self.customer_id = op["value"]
                elif "/description" in op["path"]:
                    self.description = op["value"]
                elif "/enabled" in op["path"]:
                    self.enabled = self._str2bool(op["value"])
            else:
                raise Exception(f'Patch operation "{op["op"]}" not implemented')
        return self

    def _str2bool(self, v: str) -> bool:
        return v.lower() == "true"


class UsagePlan(BaseModel):
    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        apiStages: Any = None,
        throttle: Optional[Dict[str, Any]] = None,
        quota: Optional[Dict[str, Any]] = None,
        productCode: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
    ):
        self.id = create_id()
        self.name = name
        self.description = description
        self.api_stages = apiStages or []
        self.throttle = throttle or {}
        self.quota = quota or {}
        self.product_code = productCode
        self.tags = tags

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "apiStages": self.api_stages,
            "throttle": self.throttle,
            "quota": self.quota,
            "productCode": self.product_code,
            "tags": self.tags,
        }

    def apply_patch_operations(self, patch_operations: List[Dict[str, Any]]) -> None:
        for op in patch_operations:
            path = op["path"]
            value = op["value"]
            if op["op"] == "replace":
                if "/name" in path:
                    self.name = value
                if "/productCode" in path:
                    self.product_code = value
                if "/description" in path:
                    self.description = value
                if "/quota/limit" in path:
                    self.quota["limit"] = value
                if "/quota/period" in path:
                    self.quota["period"] = value
                if "/throttle/rateLimit" in path:
                    self.throttle["rateLimit"] = value
                if "/throttle/burstLimit" in path:
                    self.throttle["burstLimit"] = value


class RequestValidator(BaseModel):
    PROP_ID = "id"
    PROP_NAME = "name"
    PROP_VALIDATE_REQUEST_BODY = "validateRequestBody"
    PROP_VALIDATE_REQUEST_PARAMETERS = "validateRequestParameters"

    # operations
    OP_PATH = "path"
    OP_VALUE = "value"
    OP_REPLACE = "replace"
    OP_OP = "op"

    def __init__(
        self,
        _id: str,
        name: str,
        validateRequestBody: Optional[bool],
        validateRequestParameters: Any,
    ):
        self.id = _id
        self.name = name
        self.validate_request_body = validateRequestBody
        self.validate_request_parameters = validateRequestParameters

    def apply_patch_operations(self, operations: List[Dict[str, Any]]) -> None:
        for operation in operations:
            path = operation[RequestValidator.OP_PATH]
            value = operation[RequestValidator.OP_VALUE]
            if operation[RequestValidator.OP_OP] == RequestValidator.OP_REPLACE:
                if to_path(RequestValidator.PROP_NAME) in path:
                    self.name = value
                if to_path(RequestValidator.PROP_VALIDATE_REQUEST_BODY) in path:
                    self.validate_request_body = value.lower() in ("true")
                if to_path(RequestValidator.PROP_VALIDATE_REQUEST_PARAMETERS) in path:
                    self.validate_request_parameters = value.lower() in ("true")

    def to_dict(self) -> Dict[str, Any]:
        return {
            RequestValidator.PROP_ID: self.id,
            RequestValidator.PROP_NAME: self.name,
            RequestValidator.PROP_VALIDATE_REQUEST_BODY: self.validate_request_body,
            RequestValidator.PROP_VALIDATE_REQUEST_PARAMETERS: self.validate_request_parameters,
        }


class UsagePlanKey(BaseModel):
    def __init__(self, plan_id: str, plan_type: str, name: Optional[str], value: str):
        self.id = plan_id
        self.name = name
        self.type = plan_type
        self.value = value

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "value": self.value,
        }


class VpcLink(BaseModel):
    def __init__(
        self,
        name: str,
        description: str,
        target_arns: List[str],
        tags: List[Dict[str, str]],
    ):
        self.id = create_id()
        self.name = name
        self.description = description
        self.target_arns = target_arns
        self.tags = tags
        self.status = "AVAILABLE"

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "targetArns": self.target_arns,
            "tags": self.tags,
            "status": self.status,
        }


class RestAPI(CloudFormationModel):

    PROP_ID = "id"
    PROP_NAME = "name"
    PROP_DESCRIPTION = "description"
    PROP_VERSION = "version"
    PROP_BINARY_MEDIA_TYPES = "binaryMediaTypes"
    PROP_CREATED_DATE = "createdDate"
    PROP_API_KEY_SOURCE = "apiKeySource"
    PROP_ENDPOINT_CONFIGURATION = "endpointConfiguration"
    PROP_TAGS = "tags"
    PROP_POLICY = "policy"
    PROP_DISABLE_EXECUTE_API_ENDPOINT = "disableExecuteApiEndpoint"
    PROP_MINIMUM_COMPRESSION_SIZE = "minimumCompressionSize"

    # operations
    OPERATION_ADD = "add"
    OPERATION_REPLACE = "replace"
    OPERATION_REMOVE = "remove"
    OPERATION_PATH = "path"
    OPERATION_VALUE = "value"
    OPERATION_OP = "op"

    def __init__(
        self,
        api_id: str,
        account_id: str,
        region_name: str,
        name: str,
        description: str,
        **kwargs: Any,
    ):
        self.id = api_id
        self.account_id = account_id
        self.region_name = region_name
        self.name = name
        self.description = description
        self.version = kwargs.get(RestAPI.PROP_VERSION) or "V1"
        self.binaryMediaTypes = kwargs.get(RestAPI.PROP_BINARY_MEDIA_TYPES) or []
        self.create_date = int(time.time())
        self.api_key_source = kwargs.get("api_key_source") or "HEADER"
        self.policy = kwargs.get(RestAPI.PROP_POLICY) or None
        self.endpoint_configuration = kwargs.get("endpoint_configuration") or {
            "types": ["EDGE"]
        }
        self.tags = kwargs.get(RestAPI.PROP_TAGS) or {}
        self.disableExecuteApiEndpoint = (
            kwargs.get("disable_execute_api_endpoint") or False
        )
        self.minimum_compression_size = kwargs.get("minimum_compression_size")
        self.deployments: Dict[str, Deployment] = {}
        self.authorizers: Dict[str, Authorizer] = {}
        self.gateway_responses: Dict[str, GatewayResponse] = {}
        self.stages: Dict[str, Stage] = {}
        self.resources: Dict[str, Resource] = {}
        self.models: Dict[str, Model] = {}
        self.request_validators: Dict[str, RequestValidator] = {}
        self.default = self.add_child("/")  # Add default child

    def __repr__(self) -> str:
        return str(self.id)

    def to_dict(self) -> Dict[str, Any]:
        return {
            self.PROP_ID: self.id,
            self.PROP_NAME: self.name,
            self.PROP_DESCRIPTION: self.description,
            self.PROP_VERSION: self.version,
            self.PROP_BINARY_MEDIA_TYPES: self.binaryMediaTypes,
            self.PROP_CREATED_DATE: self.create_date,
            self.PROP_API_KEY_SOURCE: self.api_key_source,
            self.PROP_ENDPOINT_CONFIGURATION: self.endpoint_configuration,
            self.PROP_TAGS: self.tags,
            self.PROP_POLICY: self.policy,
            self.PROP_DISABLE_EXECUTE_API_ENDPOINT: self.disableExecuteApiEndpoint,
            self.PROP_MINIMUM_COMPRESSION_SIZE: self.minimum_compression_size,
        }

    def apply_patch_operations(self, patch_operations: List[Dict[str, Any]]) -> None:
        for op in patch_operations:
            path = op[self.OPERATION_PATH]
            value = ""
            if self.OPERATION_VALUE in op:
                value = op[self.OPERATION_VALUE]
            operaton = op[self.OPERATION_OP]
            if operaton == self.OPERATION_REPLACE:
                if to_path(self.PROP_NAME) in path:
                    self.name = value
                if to_path(self.PROP_DESCRIPTION) in path:
                    self.description = value
                if to_path(self.PROP_API_KEY_SOURCE) in path:
                    self.api_key_source = value
                if to_path(self.PROP_BINARY_MEDIA_TYPES) in path:
                    self.binaryMediaTypes = [value]
                if to_path(self.PROP_DISABLE_EXECUTE_API_ENDPOINT) in path:
                    self.disableExecuteApiEndpoint = bool(value)
                if to_path(self.PROP_POLICY) in path:
                    self.policy = value
            elif operaton == self.OPERATION_ADD:
                if to_path(self.PROP_BINARY_MEDIA_TYPES) in path:
                    self.binaryMediaTypes.append(value)
            elif operaton == self.OPERATION_REMOVE:
                if to_path(self.PROP_BINARY_MEDIA_TYPES) in path:
                    self.binaryMediaTypes.remove(value)
                if to_path(self.PROP_DESCRIPTION) in path:
                    self.description = ""

    @classmethod
    def has_cfn_attr(cls, attr: str) -> bool:
        return attr in ["RootResourceId"]

    def get_cfn_attribute(self, attribute_name: str) -> Any:
        from moto.cloudformation.exceptions import UnformattedGetAttTemplateException

        if attribute_name == "RootResourceId":
            for res_id, res_obj in self.resources.items():
                if res_obj.path_part == "/" and not res_obj.parent_id:
                    return res_id
            raise Exception(f"Unable to find root resource for API {self}")
        raise UnformattedGetAttTemplateException()

    @property
    def physical_resource_id(self) -> str:
        return self.id

    @staticmethod
    def cloudformation_name_type() -> str:
        return "RestApi"

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::ApiGateway::RestApi"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Dict[str, Any],
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "RestAPI":
        properties = cloudformation_json["Properties"]
        name = properties["Name"]
        desc = properties.get("Description", "")
        config = properties.get("EndpointConfiguration", None)
        backend = apigateway_backends[account_id][region_name]
        return backend.create_rest_api(
            name=name, description=desc, endpoint_configuration=config
        )

    def add_child(self, path: str, parent_id: Optional[str] = None) -> Resource:
        child_id = create_id()
        child = Resource(
            resource_id=child_id,
            account_id=self.account_id,
            region_name=self.region_name,
            api_id=self.id,
            path_part=path,
            parent_id=parent_id,
        )
        self.resources[child_id] = child
        return child

    def add_model(
        self,
        name: str,
        description: str,
        schema: str,
        content_type: str,
    ) -> "Model":
        model_id = create_id()
        new_model = Model(
            model_id=model_id,
            name=name,
            description=description,
            schema=schema,
            content_type=content_type,
        )

        self.models[name] = new_model
        return new_model

    def get_resource_for_path(self, path_after_stage_name: str) -> Resource:  # type: ignore[return]
        for resource in self.resources.values():
            if resource.get_path() == path_after_stage_name:
                return resource
        # TODO deal with no matching resource

    def resource_callback(
        self, request: Any
    ) -> Tuple[int, Dict[str, str], Union[str, bytes]]:
        path = path_url(request.url)
        path_after_stage_name = "/" + "/".join(path.split("/")[2:])

        resource = self.get_resource_for_path(path_after_stage_name)
        status_code, response = resource.get_response(request)
        return status_code, {}, response

    def update_integration_mocks(self, stage_name: str) -> None:
        stage_url_lower = STAGE_URL.format(
            api_id=self.id.lower(), region_name=self.region_name, stage_name=stage_name
        )
        stage_url_upper = STAGE_URL.format(
            api_id=self.id.upper(), region_name=self.region_name, stage_name=stage_name
        )

        for resource in self.resources.values():
            path = resource.get_path()
            path = "" if path == "/" else path

            for http_method in resource.resource_methods.keys():
                for url in [stage_url_lower, stage_url_upper]:
                    callback_response = responses.CallbackResponse(
                        url=url + path,
                        method=http_method,
                        callback=self.resource_callback,
                        content_type="text/plain",
                    )
                    responses_mock.add(callback_response)

    def create_authorizer(
        self,
        authorizer_id: str,
        name: str,
        authorizer_type: str,
        provider_arns: Optional[List[str]],
        auth_type: Optional[str],
        authorizer_uri: Optional[str],
        authorizer_credentials: Optional[str],
        identity_source: Optional[str],
        identiy_validation_expression: Optional[str],
        authorizer_result_ttl: Optional[int],
    ) -> Authorizer:
        authorizer = Authorizer(
            authorizer_id=authorizer_id,
            name=name,
            authorizer_type=authorizer_type,
            provider_arns=provider_arns,
            auth_type=auth_type,
            authorizer_uri=authorizer_uri,
            authorizer_credentials=authorizer_credentials,
            identity_source=identity_source,
            identiy_validation_expression=identiy_validation_expression,
            authorizer_result_ttl=authorizer_result_ttl,
        )
        self.authorizers[authorizer_id] = authorizer
        return authorizer

    def create_stage(
        self,
        name: str,
        deployment_id: str,
        variables: Any,
        description: str,
        cacheClusterEnabled: Optional[bool],
        cacheClusterSize: Optional[str],
        tags: Optional[Dict[str, str]],
        tracing_enabled: Optional[bool],
    ) -> Stage:
        if name in self.stages:
            raise ConflictException("Stage already exists")
        if variables is None:
            variables = {}
        stage = Stage(
            name=name,
            deployment_id=deployment_id,
            variables=variables,
            description=description,
            cacheClusterSize=cacheClusterSize,
            cacheClusterEnabled=cacheClusterEnabled,
            tags=tags,
            tracing_enabled=tracing_enabled,
        )
        self.stages[name] = stage
        self.update_integration_mocks(name)
        return stage

    def create_deployment(
        self, name: str, description: str, stage_variables: Any = None
    ) -> Deployment:
        if stage_variables is None:
            stage_variables = {}
        deployment_id = create_id()
        deployment = Deployment(deployment_id, name, description)
        self.deployments[deployment_id] = deployment
        if name:
            self.stages[name] = Stage(
                name=name, deployment_id=deployment_id, variables=stage_variables
            )
        self.update_integration_mocks(name)

        return deployment

    def get_deployment(self, deployment_id: str) -> Deployment:
        return self.deployments[deployment_id]

    def get_authorizers(self) -> List[Authorizer]:
        return list(self.authorizers.values())

    def get_stages(self) -> List[Stage]:
        return list(self.stages.values())

    def get_deployments(self) -> List[Deployment]:
        return list(self.deployments.values())

    def delete_deployment(self, deployment_id: str) -> Deployment:
        if deployment_id not in self.deployments:
            raise DeploymentNotFoundException()
        deployment = self.deployments[deployment_id]
        if deployment.stage_name and deployment.stage_name in self.stages:
            # Stage is still active
            raise StageStillActive()

        return self.deployments.pop(deployment_id)

    def create_request_validator(
        self,
        name: str,
        validateRequestBody: Optional[bool],
        validateRequestParameters: Any,
    ) -> RequestValidator:
        validator_id = create_id()
        request_validator = RequestValidator(
            _id=validator_id,
            name=name,
            validateRequestBody=validateRequestBody,
            validateRequestParameters=validateRequestParameters,
        )
        self.request_validators[validator_id] = request_validator
        return request_validator

    def get_request_validators(self) -> List[RequestValidator]:
        return list(self.request_validators.values())

    def get_request_validator(self, validator_id: str) -> RequestValidator:
        reqeust_validator = self.request_validators.get(validator_id)
        if reqeust_validator is None:
            raise RequestValidatorNotFound()
        return reqeust_validator

    def delete_request_validator(self, validator_id: str) -> RequestValidator:
        return self.request_validators.pop(validator_id)

    def update_request_validator(
        self, validator_id: str, patch_operations: List[Dict[str, Any]]
    ) -> RequestValidator:
        self.request_validators[validator_id].apply_patch_operations(patch_operations)
        return self.request_validators[validator_id]

    def put_gateway_response(
        self,
        response_type: str,
        status_code: int,
        response_parameters: Dict[str, Any],
        response_templates: Dict[str, str],
    ) -> "GatewayResponse":
        response = GatewayResponse(
            response_type=response_type,
            status_code=status_code,
            response_parameters=response_parameters,
            response_templates=response_templates,
        )
        self.gateway_responses[response_type] = response
        return response

    def get_gateway_response(self, response_type: str) -> "GatewayResponse":
        if response_type not in self.gateway_responses:
            raise GatewayResponseNotFound()
        return self.gateway_responses[response_type]

    def get_gateway_responses(self) -> List["GatewayResponse"]:
        return list(self.gateway_responses.values())

    def delete_gateway_response(self, response_type: str) -> None:
        self.gateway_responses.pop(response_type, None)


class DomainName(BaseModel):
    def __init__(self, domain_name: str, **kwargs: Any):
        self.domain_name = domain_name
        region = kwargs.get("region_name") or "us-east-1"
        self.regional_domain_name = (
            f"d-{create_id()}.execute-api.{region}.amazonaws.com"
        )
        self.distribution_domain_name = f"d{create_id()}.cloudfront.net"
        self.domain_name_status = "AVAILABLE"
        self.status_message = "Domain Name Available"
        self.regional_hosted_zone_id = "Z2FDTNDATAQYW2"
        self.distribution_hosted_zone_id = "Z2FDTNDATAQYW2"
        self.certificate_upload_date = int(time.time())
        self.certificate_name = kwargs.get("certificate_name")
        self.certificate_arn = kwargs.get("certificate_arn")
        self.certificate_body = kwargs.get("certificate_body")
        self.tags = kwargs.get("tags")
        self.security_policy = kwargs.get("security_policy")
        self.certificate_chain = kwargs.get("certificate_chain")
        self.regional_certificate_name = kwargs.get("regional_certificate_name")
        self.certificate_private_key = kwargs.get("certificate_private_key")
        self.regional_certificate_arn = kwargs.get("regional_certificate_arn")
        self.endpoint_configuration = kwargs.get("endpoint_configuration")

    def to_json(self) -> Dict[str, Any]:
        dct = {
            "domainName": self.domain_name,
            "regionalDomainName": self.regional_domain_name,
            "distributionDomainName": self.distribution_domain_name,
            "domainNameStatus": self.domain_name_status,
            "domainNameStatusMessage": self.status_message,
            "regionalHostedZoneId": self.regional_hosted_zone_id,
            "distributionHostedZoneId": self.distribution_hosted_zone_id,
            "certificateUploadDate": self.certificate_upload_date,
        }
        if self.certificate_name:
            dct["certificateName"] = self.certificate_name
        if self.certificate_arn:
            dct["certificateArn"] = self.certificate_arn
        if self.certificate_body:
            dct["certificateBody"] = self.certificate_body
        if self.tags:
            dct["tags"] = self.tags
        if self.security_policy:
            dct["securityPolicy"] = self.security_policy
        if self.certificate_chain:
            dct["certificateChain"] = self.certificate_chain
        if self.regional_certificate_name:
            dct["regionalCertificateName"] = self.regional_certificate_name
        if self.certificate_private_key:
            dct["certificatePrivateKey"] = self.certificate_private_key
        if self.regional_certificate_arn:
            dct["regionalCertificateArn"] = self.regional_certificate_arn
        if self.endpoint_configuration:
            dct["endpointConfiguration"] = self.endpoint_configuration
        return dct


class Model(BaseModel):
    def __init__(self, model_id: str, name: str, **kwargs: Any):
        self.id = model_id
        self.name = name
        self.description = kwargs.get("description")
        self.schema = kwargs.get("schema")
        self.content_type = kwargs.get("content_type")

    def to_json(self) -> Dict[str, Any]:
        dct = {
            "id": self.id,
            "name": self.name,
        }
        if self.description:
            dct["description"] = self.description
        if self.schema:
            dct["schema"] = self.schema
        if self.content_type:
            dct["contentType"] = self.content_type
        return dct


class BasePathMapping(BaseModel):

    # operations
    OPERATION_REPLACE = "replace"
    OPERATION_PATH = "path"
    OPERATION_VALUE = "value"
    OPERATION_OP = "op"

    def __init__(self, domain_name: str, rest_api_id: str, **kwargs: Any):
        self.domain_name = domain_name
        self.rest_api_id = rest_api_id
        self.base_path = kwargs.get("basePath") or "(none)"
        self.stage = kwargs.get("stage")

    def to_json(self) -> Dict[str, Any]:
        dct = {
            "domain_name": self.domain_name,
            "restApiId": self.rest_api_id,
            "basePath": self.base_path,
        }
        if self.stage is not None:
            dct["stage"] = self.stage
        return dct

    def apply_patch_operations(self, patch_operations: List[Dict[str, Any]]) -> None:
        for op in patch_operations:
            path = op["path"]
            value = op["value"]
            operation = op["op"]
            if operation == self.OPERATION_REPLACE:
                if "/basePath" in path:
                    self.base_path = value
                if "/restapiId" in path:
                    self.rest_api_id = value
                if "/stage" in path:
                    self.stage = value


class GatewayResponse(BaseModel):
    def __init__(
        self,
        response_type: str,
        status_code: int,
        response_parameters: Dict[str, Any],
        response_templates: Dict[str, str],
    ):
        self.response_type = response_type
        self.default_response = False
        self.status_code = status_code
        self.response_parameters = response_parameters
        self.response_templates = response_templates

    def to_json(self) -> Dict[str, Any]:
        dct = {
            "responseType": self.response_type,
            "defaultResponse": self.default_response,
        }
        if self.status_code is not None:
            dct["statusCode"] = self.status_code
        if self.response_parameters is not None:
            dct["responseParameters"] = self.response_parameters
        if self.response_templates is not None:
            dct["responseTemplates"] = self.response_templates
        return dct


class APIGatewayBackend(BaseBackend):
    """
    API Gateway mock.

    The public URLs of an API integration are mocked as well, i.e. the following would be supported in Moto:

    .. sourcecode:: python

        client.put_integration(
            restApiId=api_id,
            ...,
            uri="http://httpbin.org/robots.txt",
            integrationHttpMethod="GET"
        )
        deploy_url = f"https://{api_id}.execute-api.us-east-1.amazonaws.com/dev"
        assert requests.get(deploy_url).content == b"a fake response"

    Limitations:
     - Integrations of type HTTP are supported
     - Integrations of type AWS with service DynamoDB are supported
     - Other types (AWS_PROXY, MOCK, etc) are ignored
     - Other services are not yet supported
     - The BasePath of an API is ignored
     - TemplateMapping is not yet supported for requests/responses
     - This only works when using the decorators, not in ServerMode
    """

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.apis: Dict[str, RestAPI] = {}
        self.keys: Dict[str, ApiKey] = {}
        self.usage_plans: Dict[str, UsagePlan] = {}
        self.usage_plan_keys: Dict[str, Dict[str, UsagePlanKey]] = {}
        self.domain_names: Dict[str, DomainName] = {}
        self.models: Dict[str, Model] = {}
        self.base_path_mappings: Dict[str, Dict[str, BasePathMapping]] = {}
        self.vpc_links: Dict[str, VpcLink] = {}

    def create_rest_api(
        self,
        name: str,
        description: str,
        api_key_source: Optional[str] = None,
        endpoint_configuration: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
        policy: Optional[str] = None,
        minimum_compression_size: Optional[int] = None,
        disable_execute_api_endpoint: Optional[bool] = None,
    ) -> RestAPI:
        api_id = create_id()
        rest_api = RestAPI(
            api_id,
            self.account_id,
            self.region_name,
            name,
            description,
            api_key_source=api_key_source,
            endpoint_configuration=endpoint_configuration,
            tags=tags,
            policy=policy,
            minimum_compression_size=minimum_compression_size,
            disable_execute_api_endpoint=disable_execute_api_endpoint,
        )
        self.apis[api_id] = rest_api
        return rest_api

    def import_rest_api(
        self, api_doc: Dict[str, Any], fail_on_warnings: bool
    ) -> RestAPI:
        """
        Only a subset of the OpenAPI spec 3.x is currently implemented.
        """
        if fail_on_warnings:
            try:
                validate_spec(api_doc)  # type: ignore[arg-type]
            except OpenAPIValidationError as e:
                raise InvalidOpenAPIDocumentException(e)
            except AttributeError:
                # Call can fail in Python3.7 due to `typing_extensions 4.6.0` throwing an error
                # Easiest to just ignore this for now - Py3.7 is EOL soon anyway
                pass
        name = api_doc["info"]["title"]
        description = api_doc["info"]["description"]
        api = self.create_rest_api(name=name, description=description)
        self.put_rest_api(api.id, api_doc, fail_on_warnings=fail_on_warnings)
        return api

    def export_api(self, rest_api_id: str, export_type: str) -> Dict[str, Any]:
        """
        Not all fields are implemented yet.
        The export-type is currently ignored - we will only return the 'swagger'-format
        """
        try:
            api = self.get_rest_api(rest_api_id)
        except RestAPINotFound:
            raise StageNotFoundException
        if export_type not in ["swagger", "oas30"]:
            raise BadRequestException(f"No API exporter for type '{export_type}'")
        now = datetime.now().strftime("%Y-%m-%dT%H:%m:%S")
        resp: Dict[str, Any] = {
            "swagger": "2.0",
            "info": {"version": now, "title": api.name},
            "host": f"{api.id}.execute-api.{self.region_name}.amazonaws.com",
            "basePath": "/",
            "schemes": ["https"],
            "paths": {},
            "definitions": {"Empty": {"type": "object", "title": "Empty Schema"}},
        }
        for res in api.resources.values():
            path = res.get_path()
            resp["paths"][path] = {}
            for method_type, method in res.resource_methods.items():
                resp["paths"][path][method_type] = {
                    "produces": ["application/json"],
                    "responses": {},
                }
                for code, _ in method.method_responses.items():
                    resp["paths"][path][method_type]["responses"][code] = {
                        "description": f"{code} response",
                        "schema": {"$ref": "#/definitions/Empty"},
                    }
        return resp

    def get_rest_api(self, function_id: str) -> RestAPI:
        rest_api = self.apis.get(function_id)
        if rest_api is None:
            raise RestAPINotFound()
        return rest_api

    def put_rest_api(
        self,
        function_id: str,
        api_doc: Dict[str, Any],
        fail_on_warnings: bool,
        mode: str = "merge",
    ) -> RestAPI:
        """
        Only a subset of the OpenAPI spec 3.x is currently implemented.
        """
        if mode not in ["merge", "overwrite"]:
            raise InvalidOpenApiModeException()

        if api_doc.get("swagger") is not None or (
            api_doc.get("openapi") is not None and api_doc["openapi"][0] != "3"
        ):
            raise InvalidOpenApiDocVersionException()

        if fail_on_warnings:
            try:
                validate_spec(api_doc)  # type: ignore[arg-type]
            except OpenAPIValidationError as e:
                raise InvalidOpenAPIDocumentException(e)
            except AttributeError:
                # Call can fail in Python3.7 due to `typing_extensions 4.6.0` throwing an error
                # Easiest to just ignore this for now - Py3.7 is EOL soon anyway
                pass

        if mode == "overwrite":
            api = self.get_rest_api(function_id)
            api.resources = {}
            api.default = api.add_child("/")  # Add default child

        for (path, resource_doc) in sorted(
            api_doc["paths"].items(), key=lambda x: x[0]
        ):
            # We may want to create a path like /store/inventory
            # Ensure that /store exists first, so we can use it as a parent
            ancestors = path.split("/")[
                1:-1
            ]  # skip first (empty), skip last (child) - only process ancestors
            direct_parent = ""
            parent_id = self.apis[function_id].get_resource_for_path("/").id
            for a in ancestors:
                res = self.apis[function_id].get_resource_for_path(
                    direct_parent + "/" + a
                )
                if res is None:
                    res = self.create_resource(
                        function_id=function_id,
                        parent_resource_id=parent_id,
                        path_part=a,
                    )
                parent_id = res.id
                direct_parent = direct_parent + "/" + a

            # Now that we know all ancestors are created, create the resource itself
            parent_path_part = path[0 : path.rfind("/")] or "/"
            parent_resource_id = (
                self.apis[function_id].get_resource_for_path(parent_path_part).id
            )
            resource = self.create_resource(
                function_id=function_id,
                parent_resource_id=parent_resource_id,
                path_part=path[path.rfind("/") + 1 :],
            )

            for (method_type, method_doc) in resource_doc.items():
                method_type = method_type.upper()
                if method_doc.get("x-amazon-apigateway-integration") is None:
                    self.put_method(function_id, resource.id, method_type, None)
                    method_responses = method_doc.get("responses", {}).items()
                    for (response_code, _) in method_responses:
                        self.put_method_response(
                            function_id,
                            resource.id,
                            method_type,
                            response_code,
                            response_models=None,
                            response_parameters=None,
                        )

        return self.get_rest_api(function_id)

    def update_rest_api(
        self, function_id: str, patch_operations: List[Dict[str, Any]]
    ) -> RestAPI:
        rest_api = self.apis.get(function_id)
        if rest_api is None:
            raise RestAPINotFound()
        self.apis[function_id].apply_patch_operations(patch_operations)
        return self.apis[function_id]

    def list_apis(self) -> List[RestAPI]:
        return list(self.apis.values())

    def delete_rest_api(self, function_id: str) -> RestAPI:
        rest_api = self.apis.pop(function_id)
        return rest_api

    def get_resources(self, function_id: str) -> List[Resource]:
        api = self.get_rest_api(function_id)
        return list(api.resources.values())

    def get_resource(self, function_id: str, resource_id: str) -> Resource:
        api = self.get_rest_api(function_id)
        if resource_id not in api.resources:
            raise ResourceIdNotFoundException
        return api.resources[resource_id]

    def create_resource(
        self, function_id: str, parent_resource_id: str, path_part: str
    ) -> Resource:
        api = self.get_rest_api(function_id)
        if not path_part:
            # We're attempting to create the default resource, which already exists.
            return api.default
        if not re.match("^\\{?[a-zA-Z0-9._\\-\\:]+\\+?\\}?$", path_part):
            raise InvalidResourcePathException()
        return api.add_child(path=path_part, parent_id=parent_resource_id)

    def delete_resource(self, function_id: str, resource_id: str) -> Resource:
        api = self.get_rest_api(function_id)
        return api.resources.pop(resource_id)

    def get_method(
        self, function_id: str, resource_id: str, method_type: str
    ) -> Method:
        resource = self.get_resource(function_id, resource_id)
        return resource.get_method(method_type)

    def put_method(
        self,
        function_id: str,
        resource_id: str,
        method_type: str,
        authorization_type: Optional[str],
        api_key_required: Optional[bool] = None,
        request_parameters: Optional[Dict[str, Any]] = None,
        request_models: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        authorizer_id: Optional[str] = None,
        authorization_scopes: Optional[str] = None,
        request_validator_id: Optional[str] = None,
    ) -> Method:
        resource = self.get_resource(function_id, resource_id)
        method = resource.add_method(
            method_type,
            authorization_type,
            api_key_required=api_key_required,
            request_parameters=request_parameters,
            request_models=request_models,
            operation_name=operation_name,
            authorizer_id=authorizer_id,
            authorization_scopes=authorization_scopes,
            request_validator_id=request_validator_id,
        )
        return method

    def delete_method(
        self, function_id: str, resource_id: str, method_type: str
    ) -> None:
        resource = self.get_resource(function_id, resource_id)
        resource.delete_method(method_type)

    def get_authorizer(self, restapi_id: str, authorizer_id: str) -> Authorizer:
        api = self.get_rest_api(restapi_id)
        authorizer = api.authorizers.get(authorizer_id)
        if authorizer is None:
            raise AuthorizerNotFoundException()
        else:
            return authorizer

    def get_authorizers(self, restapi_id: str) -> List[Authorizer]:
        api = self.get_rest_api(restapi_id)
        return api.get_authorizers()

    def create_authorizer(
        self, restapi_id: str, name: str, authorizer_type: str, **kwargs: Any
    ) -> Authorizer:
        api = self.get_rest_api(restapi_id)
        authorizer_id = create_id()
        return api.create_authorizer(
            authorizer_id,
            name,
            authorizer_type,
            provider_arns=kwargs.get("provider_arns"),
            auth_type=kwargs.get("auth_type"),
            authorizer_uri=kwargs.get("authorizer_uri"),
            authorizer_credentials=kwargs.get("authorizer_credentials"),
            identity_source=kwargs.get("identity_source"),
            identiy_validation_expression=kwargs.get("identiy_validation_expression"),
            authorizer_result_ttl=kwargs.get("authorizer_result_ttl"),
        )

    def update_authorizer(
        self, restapi_id: str, authorizer_id: str, patch_operations: Any
    ) -> Authorizer:
        authorizer = self.get_authorizer(restapi_id, authorizer_id)
        return authorizer.apply_operations(patch_operations)

    def delete_authorizer(self, restapi_id: str, authorizer_id: str) -> None:
        api = self.get_rest_api(restapi_id)
        del api.authorizers[authorizer_id]

    def get_stage(self, function_id: str, stage_name: str) -> Stage:
        api = self.get_rest_api(function_id)
        stage = api.stages.get(stage_name)
        if stage is None:
            raise StageNotFoundException()
        return stage

    def get_stages(self, function_id: str) -> List[Stage]:
        api = self.get_rest_api(function_id)
        return api.get_stages()

    def create_stage(
        self,
        function_id: str,
        stage_name: str,
        deploymentId: str,
        variables: Optional[Any] = None,
        description: str = "",
        cacheClusterEnabled: Optional[bool] = None,
        cacheClusterSize: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        tracing_enabled: Optional[bool] = None,
    ) -> Stage:
        if variables is None:
            variables = {}
        api = self.get_rest_api(function_id)
        return api.create_stage(
            stage_name,
            deploymentId,
            variables=variables,
            description=description,
            cacheClusterEnabled=cacheClusterEnabled,
            cacheClusterSize=cacheClusterSize,
            tags=tags,
            tracing_enabled=tracing_enabled,
        )

    def update_stage(
        self, function_id: str, stage_name: str, patch_operations: Any
    ) -> Stage:
        stage = self.get_stage(function_id, stage_name)
        if not stage:
            api = self.get_rest_api(function_id)
            stage = api.stages[stage_name] = Stage()
        return stage.apply_operations(patch_operations)

    def delete_stage(self, function_id: str, stage_name: str) -> None:
        api = self.get_rest_api(function_id)
        deleted = api.stages.pop(stage_name, None)
        if not deleted:
            raise StageNotFoundException()

    def get_method_response(
        self, function_id: str, resource_id: str, method_type: str, response_code: str
    ) -> Optional[MethodResponse]:
        method = self.get_method(function_id, resource_id, method_type)
        return method.get_response(response_code)

    def put_method_response(
        self,
        function_id: str,
        resource_id: str,
        method_type: str,
        response_code: str,
        response_models: Any,
        response_parameters: Any,
    ) -> MethodResponse:
        method = self.get_method(function_id, resource_id, method_type)
        return method.create_response(
            response_code, response_models, response_parameters
        )

    def delete_method_response(
        self, function_id: str, resource_id: str, method_type: str, response_code: str
    ) -> Optional[MethodResponse]:
        method = self.get_method(function_id, resource_id, method_type)
        return method.delete_response(response_code)

    def put_integration(
        self,
        function_id: str,
        resource_id: str,
        method_type: str,
        integration_type: str,
        uri: str,
        integration_method: str,
        credentials: Optional[str] = None,
        request_templates: Optional[Dict[str, Any]] = None,
        passthrough_behavior: Optional[str] = None,
        tls_config: Optional[Dict[str, Any]] = None,
        cache_namespace: Optional[str] = None,
        timeout_in_millis: Optional[str] = None,
        request_parameters: Optional[Dict[str, Any]] = None,
        content_handling: Optional[str] = None,
        connection_type: Optional[str] = None,
    ) -> Integration:
        resource = self.get_resource(function_id, resource_id)
        if credentials and not re.match(
            "^arn:aws:iam::" + str(self.account_id), credentials
        ):
            raise CrossAccountNotAllowed()
        if not integration_method and integration_type in [
            "HTTP",
            "HTTP_PROXY",
            "AWS",
            "AWS_PROXY",
        ]:
            raise IntegrationMethodNotDefined()
        if integration_type in ["AWS_PROXY"] and re.match(
            "^arn:aws:apigateway:[a-zA-Z0-9-]+:s3", uri
        ):
            raise AwsProxyNotAllowed()
        if (
            integration_type in ["AWS"]
            and re.match("^arn:aws:apigateway:[a-zA-Z0-9-]+:s3", uri)
            and not credentials
        ):
            raise RoleNotSpecified()
        if integration_type in ["HTTP", "HTTP_PROXY"] and not self._uri_validator(uri):
            raise InvalidHttpEndpoint()
        if integration_type in ["AWS", "AWS_PROXY"] and not re.match("^arn:aws:", uri):
            raise InvalidArn()
        if integration_type in ["AWS", "AWS_PROXY"] and not re.match(
            "^arn:aws:apigateway:[a-zA-Z0-9-]+:[a-zA-Z0-9-.]+:(path|action)/", uri
        ):
            raise InvalidIntegrationArn()
        integration = resource.add_integration(
            method_type,
            integration_type,
            uri,
            integration_method=integration_method,
            request_templates=request_templates,
            passthrough_behavior=passthrough_behavior,
            tls_config=tls_config,
            cache_namespace=cache_namespace,
            timeout_in_millis=timeout_in_millis,
            request_parameters=request_parameters,
            content_handling=content_handling,
            credentials=credentials,
            connection_type=connection_type,
        )
        return integration

    def get_integration(
        self, function_id: str, resource_id: str, method_type: str
    ) -> Integration:
        resource = self.get_resource(function_id, resource_id)
        return resource.get_integration(method_type)  # type: ignore[return-value]

    def delete_integration(
        self, function_id: str, resource_id: str, method_type: str
    ) -> Integration:
        resource = self.get_resource(function_id, resource_id)
        return resource.delete_integration(method_type)

    def put_integration_response(
        self,
        function_id: str,
        resource_id: str,
        method_type: str,
        status_code: str,
        selection_pattern: str,
        response_templates: Dict[str, str],
        response_parameters: Dict[str, str],
        content_handling: str,
    ) -> IntegrationResponse:
        integration = self.get_integration(function_id, resource_id, method_type)
        if integration:
            return integration.create_integration_response(
                status_code,
                selection_pattern,
                response_templates,
                response_parameters,
                content_handling,
            )
        raise NoIntegrationResponseDefined()

    def get_integration_response(
        self, function_id: str, resource_id: str, method_type: str, status_code: str
    ) -> IntegrationResponse:
        integration = self.get_integration(function_id, resource_id, method_type)
        integration_response = integration.get_integration_response(status_code)
        return integration_response

    def delete_integration_response(
        self, function_id: str, resource_id: str, method_type: str, status_code: str
    ) -> IntegrationResponse:
        integration = self.get_integration(function_id, resource_id, method_type)
        integration_response = integration.delete_integration_response(status_code)
        return integration_response

    def create_deployment(
        self,
        function_id: str,
        name: str,
        description: str = "",
        stage_variables: Any = None,
    ) -> Deployment:
        if stage_variables is None:
            stage_variables = {}
        api = self.get_rest_api(function_id)
        nested_methods = [
            list(res.resource_methods.values())
            for res in self.get_resources(function_id)
        ]
        methods = [m for sublist in nested_methods for m in sublist]
        if not any(methods):
            raise NoMethodDefined()
        method_integrations = [
            method.method_integration for method in methods if method.method_integration
        ]
        if not any(method_integrations):
            raise NoIntegrationDefined()
        deployment = api.create_deployment(name, description, stage_variables)
        return deployment

    def get_deployment(self, function_id: str, deployment_id: str) -> Deployment:
        api = self.get_rest_api(function_id)
        return api.get_deployment(deployment_id)

    def get_deployments(self, function_id: str) -> List[Deployment]:
        api = self.get_rest_api(function_id)
        return api.get_deployments()

    def delete_deployment(self, function_id: str, deployment_id: str) -> Deployment:
        api = self.get_rest_api(function_id)
        return api.delete_deployment(deployment_id)

    def create_api_key(self, payload: Dict[str, Any]) -> ApiKey:
        if payload.get("value"):
            if len(payload.get("value", [])) < 20:
                raise ApiKeyValueMinLength()
            for api_key in self.get_api_keys():
                if api_key.value == payload["value"]:
                    raise ApiKeyAlreadyExists()
        key = ApiKey(**payload)
        self.keys[key.id] = key
        return key

    def get_api_keys(self) -> List[ApiKey]:
        return list(self.keys.values())

    def get_api_key(self, api_key_id: str) -> ApiKey:
        if api_key_id not in self.keys:
            raise ApiKeyNotFoundException()
        return self.keys[api_key_id]

    def update_api_key(self, api_key_id: str, patch_operations: Any) -> ApiKey:
        key = self.keys[api_key_id]
        return key.update_operations(patch_operations)

    def delete_api_key(self, api_key_id: str) -> None:
        self.keys.pop(api_key_id)

    def create_usage_plan(self, payload: Any) -> UsagePlan:
        plan = UsagePlan(**payload)
        self.usage_plans[plan.id] = plan
        return plan

    def get_usage_plans(self, api_key_id: Optional[str] = None) -> List[UsagePlan]:
        plans = list(self.usage_plans.values())
        if api_key_id is not None:
            plans = [
                plan
                for plan in plans
                if dict(self.usage_plan_keys.get(plan.id, {})).get(api_key_id)
            ]
        return plans

    def get_usage_plan(self, usage_plan_id: str) -> UsagePlan:
        if usage_plan_id not in self.usage_plans:
            raise UsagePlanNotFoundException()
        return self.usage_plans[usage_plan_id]

    def update_usage_plan(self, usage_plan_id: str, patch_operations: Any) -> UsagePlan:
        if usage_plan_id not in self.usage_plans:
            raise UsagePlanNotFoundException()
        self.usage_plans[usage_plan_id].apply_patch_operations(patch_operations)
        return self.usage_plans[usage_plan_id]

    def delete_usage_plan(self, usage_plan_id: str) -> None:
        self.usage_plans.pop(usage_plan_id)

    def create_usage_plan_key(
        self, usage_plan_id: str, payload: Dict[str, Any]
    ) -> UsagePlanKey:
        if usage_plan_id not in self.usage_plan_keys:
            self.usage_plan_keys[usage_plan_id] = {}

        key_id = payload["keyId"]
        if key_id not in self.keys:
            raise ApiKeyNotFoundException()

        api_key = self.keys[key_id]

        usage_plan_key = UsagePlanKey(
            plan_id=key_id,
            plan_type=payload["keyType"],
            name=api_key.name,
            value=api_key.value,
        )
        self.usage_plan_keys[usage_plan_id][usage_plan_key.id] = usage_plan_key
        return usage_plan_key

    def get_usage_plan_keys(self, usage_plan_id: str) -> List[UsagePlanKey]:
        if usage_plan_id not in self.usage_plan_keys:
            return []

        return list(self.usage_plan_keys[usage_plan_id].values())

    def get_usage_plan_key(self, usage_plan_id: str, key_id: str) -> UsagePlanKey:
        # first check if is a valid api key
        if key_id not in self.keys:
            raise ApiKeyNotFoundException()

        # then check if is a valid api key and that the key is in the plan
        if (
            usage_plan_id not in self.usage_plan_keys
            or key_id not in self.usage_plan_keys[usage_plan_id]
        ):
            raise UsagePlanNotFoundException()

        return self.usage_plan_keys[usage_plan_id][key_id]

    def delete_usage_plan_key(self, usage_plan_id: str, key_id: str) -> None:
        self.usage_plan_keys[usage_plan_id].pop(key_id)

    def _uri_validator(self, uri: str) -> bool:
        try:
            result = urlparse(uri)
            return all([result.scheme, result.netloc, result.path or "/"])
        except Exception:
            return False

    def create_domain_name(
        self,
        domain_name: str,
        certificate_name: str,
        tags: List[Dict[str, str]],
        certificate_arn: str,
        certificate_body: str,
        certificate_private_key: str,
        certificate_chain: str,
        regional_certificate_name: str,
        regional_certificate_arn: str,
        endpoint_configuration: Any,
        security_policy: str,
    ) -> DomainName:
        if not domain_name:
            raise InvalidDomainName()

        new_domain_name = DomainName(
            domain_name=domain_name,
            certificate_name=certificate_name,
            certificate_private_key=certificate_private_key,
            certificate_arn=certificate_arn,
            certificate_body=certificate_body,
            certificate_chain=certificate_chain,
            regional_certificate_name=regional_certificate_name,
            regional_certificate_arn=regional_certificate_arn,
            endpoint_configuration=endpoint_configuration,
            tags=tags,
            security_policy=security_policy,
            region_name=self.region_name,
        )

        self.domain_names[domain_name] = new_domain_name
        return new_domain_name

    def get_domain_names(self) -> List[DomainName]:
        return list(self.domain_names.values())

    def get_domain_name(self, domain_name: str) -> DomainName:
        domain_info = self.domain_names.get(domain_name)
        if domain_info is None:
            raise DomainNameNotFound()
        else:
            return domain_info

    def delete_domain_name(self, domain_name: str) -> None:
        domain_info = self.domain_names.pop(domain_name, None)
        if domain_info is None:
            raise DomainNameNotFound()

    def create_model(
        self,
        rest_api_id: str,
        name: str,
        content_type: str,
        description: str,
        schema: str,
    ) -> Model:
        if not rest_api_id:
            raise InvalidRestApiId
        if not name:
            raise InvalidModelName

        api = self.get_rest_api(rest_api_id)
        new_model = api.add_model(
            name=name,
            description=description,
            schema=schema,
            content_type=content_type,
        )

        return new_model

    def get_models(self, rest_api_id: str) -> List[Model]:
        if not rest_api_id:
            raise InvalidRestApiId
        api = self.get_rest_api(rest_api_id)
        models = api.models.values()
        return list(models)

    def get_model(self, rest_api_id: str, model_name: str) -> Model:
        if not rest_api_id:
            raise InvalidRestApiId
        api = self.get_rest_api(rest_api_id)
        model = api.models.get(model_name)
        if model is None:
            raise ModelNotFound
        else:
            return model

    def get_request_validators(self, restapi_id: str) -> List[RequestValidator]:
        restApi = self.get_rest_api(restapi_id)
        return restApi.get_request_validators()

    def create_request_validator(
        self, restapi_id: str, name: str, body: Optional[bool], params: Any
    ) -> RequestValidator:
        restApi = self.get_rest_api(restapi_id)
        return restApi.create_request_validator(
            name=name, validateRequestBody=body, validateRequestParameters=params
        )

    def get_request_validator(
        self, restapi_id: str, validator_id: str
    ) -> RequestValidator:
        restApi = self.get_rest_api(restapi_id)
        return restApi.get_request_validator(validator_id)

    def delete_request_validator(self, restapi_id: str, validator_id: str) -> None:
        restApi = self.get_rest_api(restapi_id)
        restApi.delete_request_validator(validator_id)

    def update_request_validator(
        self, restapi_id: str, validator_id: str, patch_operations: Any
    ) -> RequestValidator:
        restApi = self.get_rest_api(restapi_id)
        return restApi.update_request_validator(validator_id, patch_operations)

    def create_base_path_mapping(
        self, domain_name: str, rest_api_id: str, base_path: str, stage: str
    ) -> BasePathMapping:
        if domain_name not in self.domain_names:
            raise DomainNameNotFound()

        if base_path and "/" in base_path:
            raise InvalidBasePathException()

        if rest_api_id not in self.apis:
            raise InvalidRestApiIdForBasePathMappingException()

        if stage and self.apis[rest_api_id].stages.get(stage) is None:
            raise InvalidStageException()

        new_base_path_mapping = BasePathMapping(
            domain_name=domain_name,
            rest_api_id=rest_api_id,
            basePath=base_path,
            stage=stage,
        )

        new_base_path = new_base_path_mapping.base_path
        if self.base_path_mappings.get(domain_name) is None:
            self.base_path_mappings[domain_name] = {}
        else:
            if (
                self.base_path_mappings[domain_name].get(new_base_path)
                and new_base_path != "(none)"
            ):
                raise BasePathConflictException()
        self.base_path_mappings[domain_name][new_base_path] = new_base_path_mapping
        return new_base_path_mapping

    def get_base_path_mappings(self, domain_name: str) -> List[BasePathMapping]:
        if domain_name not in self.domain_names:
            raise DomainNameNotFound()

        return list(self.base_path_mappings[domain_name].values())

    def get_base_path_mapping(
        self, domain_name: str, base_path: str
    ) -> BasePathMapping:
        if domain_name not in self.domain_names:
            raise DomainNameNotFound()

        if base_path not in self.base_path_mappings[domain_name]:
            raise BasePathNotFoundException()

        return self.base_path_mappings[domain_name][base_path]

    def delete_base_path_mapping(self, domain_name: str, base_path: str) -> None:
        if domain_name not in self.domain_names:
            raise DomainNameNotFound()

        if base_path not in self.base_path_mappings[domain_name]:
            raise BasePathNotFoundException()

        self.base_path_mappings[domain_name].pop(base_path)

    def update_base_path_mapping(
        self, domain_name: str, base_path: str, patch_operations: Any
    ) -> BasePathMapping:

        if domain_name not in self.domain_names:
            raise DomainNameNotFound()

        if base_path not in self.base_path_mappings[domain_name]:
            raise BasePathNotFoundException()

        base_path_mapping = self.get_base_path_mapping(domain_name, base_path)

        rest_api_ids = [
            op["value"] for op in patch_operations if op["path"] == "/restapiId"
        ]
        if len(rest_api_ids) == 0:
            modified_rest_api_id = base_path_mapping.rest_api_id
        else:
            modified_rest_api_id = rest_api_ids[-1]

        stages = [op["value"] for op in patch_operations if op["path"] == "/stage"]
        if len(stages) == 0:
            modified_stage = base_path_mapping.stage
        else:
            modified_stage = stages[-1]

        base_paths = [
            op["value"] for op in patch_operations if op["path"] == "/basePath"
        ]
        if len(base_paths) == 0:
            modified_base_path = base_path_mapping.base_path
        else:
            modified_base_path = base_paths[-1]

        rest_api = self.apis.get(modified_rest_api_id)
        if rest_api is None:
            raise InvalidRestApiIdForBasePathMappingException()
        if modified_stage and rest_api.stages.get(modified_stage) is None:
            raise InvalidStageException()

        base_path_mapping.apply_patch_operations(patch_operations)

        if base_path != modified_base_path:
            self.base_path_mappings[domain_name].pop(base_path)
            self.base_path_mappings[domain_name][modified_base_path] = base_path_mapping

        return base_path_mapping

    def create_vpc_link(
        self,
        name: str,
        description: str,
        target_arns: List[str],
        tags: List[Dict[str, str]],
    ) -> VpcLink:
        vpc_link = VpcLink(
            name, description=description, target_arns=target_arns, tags=tags
        )
        self.vpc_links[vpc_link.id] = vpc_link
        return vpc_link

    def delete_vpc_link(self, vpc_link_id: str) -> None:
        self.vpc_links.pop(vpc_link_id, None)

    def get_vpc_link(self, vpc_link_id: str) -> VpcLink:
        if vpc_link_id not in self.vpc_links:
            raise VpcLinkNotFound
        return self.vpc_links[vpc_link_id]

    def get_vpc_links(self) -> List[VpcLink]:
        """
        Pagination has not yet been implemented
        """
        return list(self.vpc_links.values())

    def put_gateway_response(
        self,
        rest_api_id: str,
        response_type: str,
        status_code: int,
        response_parameters: Any,
        response_templates: Any,
    ) -> GatewayResponse:
        api = self.get_rest_api(rest_api_id)
        response = api.put_gateway_response(
            response_type,
            status_code=status_code,
            response_parameters=response_parameters,
            response_templates=response_templates,
        )
        return response

    def get_gateway_response(
        self, rest_api_id: str, response_type: str
    ) -> GatewayResponse:
        api = self.get_rest_api(rest_api_id)
        return api.get_gateway_response(response_type)

    def get_gateway_responses(self, rest_api_id: str) -> List[GatewayResponse]:
        """
        Pagination is not yet implemented
        """
        api = self.get_rest_api(rest_api_id)
        return api.get_gateway_responses()

    def delete_gateway_response(self, rest_api_id: str, response_type: str) -> None:
        api = self.get_rest_api(rest_api_id)
        api.delete_gateway_response(response_type)


apigateway_backends = BackendDict(APIGatewayBackend, "apigateway")



from typing import Any, Dict, Optional

from .exceptions import MissingRenderingAttributeException
from moto.utilities.tokenizer import GenericTokenizer


def parse_template(
    template: str,
    template_data: Dict[str, Any],
    tokenizer: Optional[GenericTokenizer] = None,
    until: Optional[str] = None,
) -> str:
    tokenizer = tokenizer or GenericTokenizer(template)
    state = ""
    parsed = ""
    var_name = ""
    for char in tokenizer:
        if until is not None and (char + tokenizer.peek(len(until) - 1)) == until:
            return parsed
        if char == "{" and tokenizer.peek() == "{":
            # Two braces next to each other indicate a variable/language construct such as for-each
            tokenizer.skip_characters("{")
            tokenizer.skip_white_space()
            if tokenizer.peek() == "#":
                state = "LOOP"
                tokenizer.skip_characters("#each")
                tokenizer.skip_white_space()
            else:
                state = "VAR"
            continue
        if char == "}":
            if state in ["LOOP", "VAR"]:
                tokenizer.skip_characters("}")
                if state == "VAR":
                    if template_data.get(var_name) is None:
                        raise MissingRenderingAttributeException(var_name)
                    parsed += template_data.get(var_name)  # type: ignore
                else:
                    current_position = tokenizer.token_pos
                    for item in template_data.get(var_name, []):
                        tokenizer.token_pos = current_position
                        parsed += parse_template(
                            template, item, tokenizer, until="{{/each}}"
                        )
                        tokenizer.skip_characters("{/each}}")
                var_name = ""
                state = ""
                continue
        if state in ["LOOP", "VAR"]:
            var_name += char
            continue
        parsed += char
    return parsed



import ipaddress
import itertools
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Tuple, TYPE_CHECKING, Set

from moto.core import CloudFormationModel

if TYPE_CHECKING:
    from moto.ec2.models.instances import Instance
from moto.ec2.models.availability_zones_and_regions import Zone
from ..exceptions import (
    GenericInvalidParameterValueError,
    InvalidAvailabilityZoneError,
    InvalidCIDRBlockParameterError,
    InvalidParameterValueError,
    InvalidSubnetConflictError,
    InvalidSubnetIdError,
    InvalidSubnetRangeError,
    InvalidSubnetCidrBlockAssociationID,
)
from .core import TaggedEC2Resource
from .availability_zones_and_regions import RegionsAndZonesBackend
from ..utils import (
    random_subnet_id,
    generic_filter,
    random_subnet_ipv6_cidr_block_association_id,
)


class Subnet(TaggedEC2Resource, CloudFormationModel):
    def __init__(
        self,
        ec2_backend: Any,
        subnet_id: str,
        vpc_id: str,
        cidr_block: str,
        ipv6_cidr_block: Optional[str],
        availability_zone: Zone,
        default_for_az: str,
        map_public_ip_on_launch: str,
        assign_ipv6_address_on_creation: bool = False,
    ):
        self.ec2_backend = ec2_backend
        self.id = subnet_id
        self.vpc_id = vpc_id
        self.cidr_block = cidr_block
        self.cidr = ipaddress.IPv4Network(str(self.cidr_block), strict=False)
        self._available_ip_addresses = self.cidr.num_addresses - 5
        self._availability_zone = availability_zone
        self.default_for_az = default_for_az
        self.map_public_ip_on_launch = map_public_ip_on_launch
        self.assign_ipv6_address_on_creation = assign_ipv6_address_on_creation
        self.ipv6_cidr_block_associations: Dict[str, Dict[str, Any]] = {}
        if ipv6_cidr_block:
            self.attach_ipv6_cidr_block_associations(ipv6_cidr_block)

        # Theory is we assign ip's as we go (as 16,777,214 usable IPs in a /8)
        self._subnet_ip_generator = self.cidr.hosts()
        self.reserved_ips = [
            next(self._subnet_ip_generator) for _ in range(0, 3)
        ]  # Reserved by AWS
        self._unused_ips: Set[
            str
        ] = set()  # if instance is destroyed hold IP here for reuse
        self._subnet_ips: Dict[str, "Instance"] = {}
        self.state = "available"

        # Placeholder for response templates until Ipv6 support implemented.
        self.ipv6_native = False

    @property
    def owner_id(self) -> str:
        return self.ec2_backend.account_id

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-subnet.html
        return "AWS::EC2::Subnet"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "Subnet":
        from ..models import ec2_backends

        properties = cloudformation_json["Properties"]

        vpc_id = properties["VpcId"]
        cidr_block = properties["CidrBlock"]
        availability_zone = properties.get("AvailabilityZone")
        ec2_backend = ec2_backends[account_id][region_name]
        subnet = ec2_backend.create_subnet(
            vpc_id=vpc_id, cidr_block=cidr_block, availability_zone=availability_zone
        )
        for tag in properties.get("Tags", []):
            tag_key = tag["Key"]
            tag_value = tag["Value"]
            subnet.add_tag(tag_key, tag_value)

        return subnet

    @classmethod
    def delete_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Dict[str, Any],
        account_id: str,
        region_name: str,
    ) -> None:
        from ..models import ec2_backends

        ec2_backends[account_id][region_name].delete_subnet(resource_name)

    @property
    def available_ip_addresses(self) -> str:
        enis = [
            eni
            for eni in self.ec2_backend.get_all_network_interfaces()
            if eni.subnet.id == self.id
        ]
        addresses_taken = []
        for eni in enis:
            if eni.private_ip_addresses:
                addresses_taken.extend(eni.private_ip_addresses)
        return str(self._available_ip_addresses - len(addresses_taken))

    @property
    def availability_zone(self) -> str:
        return self._availability_zone.name

    @property
    def availability_zone_id(self) -> str:
        return self._availability_zone.zone_id

    @property
    def physical_resource_id(self) -> str:
        return self.id

    def get_filter_value(
        self, filter_name: str, method_name: Optional[str] = None
    ) -> Any:
        """
        API Version 2014-10-01 defines the following filters for DescribeSubnets:

        * availabilityZone
        * available-ip-address-count
        * cidrBlock
        * defaultForAz
        * state
        * subnet-id
        * tag:key=value
        * tag-key
        * tag-value
        * vpc-id

        Taken from: http://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeSubnets.html
        """
        if filter_name in ("cidr", "cidrBlock", "cidr-block"):
            return self.cidr_block
        elif filter_name in ("vpc-id", "vpcId"):
            return self.vpc_id
        elif filter_name == "subnet-id":
            return self.id
        elif filter_name in ("availabilityZone", "availability-zone"):
            return self.availability_zone
        elif filter_name in ("defaultForAz", "default-for-az"):
            return self.default_for_az
        elif filter_name == "state":
            return self.state
        else:
            return super().get_filter_value(filter_name, "DescribeSubnets")

    @classmethod
    def has_cfn_attr(cls, attr: str) -> bool:
        return attr in ["AvailabilityZone"]

    def get_cfn_attribute(self, attribute_name: str) -> None:
        from moto.cloudformation.exceptions import UnformattedGetAttTemplateException

        if attribute_name == "AvailabilityZone":
            raise NotImplementedError('"Fn::GetAtt" : [ "{0}" , "AvailabilityZone" ]"')
        raise UnformattedGetAttTemplateException()

    def get_available_subnet_ip(self, instance: "Instance") -> str:
        try:
            new_ip = str(self._unused_ips.pop())
        except KeyError:
            new_ip_v4 = next(self._subnet_ip_generator)

            # Skips any IP's if they've been manually specified
            while str(new_ip_v4) in self._subnet_ips:
                new_ip_v4 = next(self._subnet_ip_generator)

            if new_ip_v4 == self.cidr.broadcast_address:
                raise StopIteration()  # Broadcast address cant be used obviously
            new_ip = str(new_ip_v4)
        # TODO StopIteration will be raised if no ip's available, not sure how aws handles this.

        self._subnet_ips[new_ip] = instance

        return new_ip

    # EFS calls this method as request_ip(str, MountTarget)
    # So technically it's not just Instances that are stored
    def request_ip(self, ip: str, instance: "Instance") -> None:
        if ipaddress.ip_address(ip) not in self.cidr:
            raise Exception(f"IP does not fall in the subnet CIDR of {self.cidr}")

        if ip in self._subnet_ips:
            raise Exception("IP already in use")
        try:
            self._unused_ips.remove(ip)
        except KeyError:
            pass

        self._subnet_ips[ip] = instance

    def del_subnet_ip(self, ip: str) -> None:
        try:
            del self._subnet_ips[ip]
            self._unused_ips.add(ip)
        except KeyError:
            pass  # Unknown IP

    def attach_ipv6_cidr_block_associations(
        self, ipv6_cidr_block: str
    ) -> Dict[str, str]:
        association = {
            "associationId": random_subnet_ipv6_cidr_block_association_id(),
            "ipv6CidrBlock": ipv6_cidr_block,
            "ipv6CidrBlockState": "associated",
        }
        self.ipv6_cidr_block_associations[association["associationId"]] = association
        return association

    def detach_subnet_cidr_block(self, association_id: str) -> Dict[str, Any]:
        association = self.ipv6_cidr_block_associations.get(association_id)
        association["ipv6CidrBlockState"] = "disassociated"  # type: ignore[index]
        return association  # type: ignore[return-value]


class SubnetRouteTableAssociation(CloudFormationModel):
    def __init__(self, route_table_id: str, subnet_id: str):
        self.route_table_id = route_table_id
        self.subnet_id = subnet_id

    @property
    def physical_resource_id(self) -> str:
        return self.route_table_id

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-subnetroutetableassociation.html
        return "AWS::EC2::SubnetRouteTableAssociation"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "SubnetRouteTableAssociation":
        from ..models import ec2_backends

        properties = cloudformation_json["Properties"]

        route_table_id = properties["RouteTableId"]
        subnet_id = properties["SubnetId"]

        ec2_backend = ec2_backends[account_id][region_name]
        subnet_association = ec2_backend.create_subnet_association(
            route_table_id=route_table_id, subnet_id=subnet_id
        )
        return subnet_association


class SubnetBackend:
    def __init__(self) -> None:
        # maps availability zone to dict of (subnet_id, subnet)
        self.subnets: Dict[str, Dict[str, Subnet]] = defaultdict(dict)
        self.subnet_associations: Dict[str, SubnetRouteTableAssociation] = {}

    def get_subnet(self, subnet_id: str) -> Subnet:
        for subnets_per_zone in self.subnets.values():
            if subnet_id in subnets_per_zone:
                return subnets_per_zone[subnet_id]
        raise InvalidSubnetIdError(subnet_id)

    def get_default_subnet(self, availability_zone: str) -> Subnet:
        return [
            subnet
            for subnet in self.describe_subnets(
                filters={"availabilityZone": availability_zone}
            )
            if subnet.default_for_az
        ][0]

    def create_subnet(
        self,
        vpc_id: str,
        cidr_block: str,
        ipv6_cidr_block: Optional[str] = None,
        availability_zone: Optional[str] = None,
        availability_zone_id: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
    ) -> Subnet:
        subnet_id = random_subnet_id()
        # Validate VPC exists and the supplied CIDR block is a subnet of the VPC's
        vpc = self.get_vpc(vpc_id)  # type: ignore[attr-defined]
        vpc_cidr_blocks = [
            ipaddress.IPv4Network(
                str(cidr_block_association["cidr_block"]), strict=False
            )
            for cidr_block_association in vpc.get_cidr_block_association_set()
        ]
        try:
            subnet_cidr_block = ipaddress.IPv4Network(str(cidr_block), strict=False)
        except ValueError:
            raise InvalidCIDRBlockParameterError(cidr_block)

        subnet_in_vpc_cidr_range = False
        for vpc_cidr_block in vpc_cidr_blocks:
            if (
                vpc_cidr_block.network_address <= subnet_cidr_block.network_address
                and vpc_cidr_block.broadcast_address
                >= subnet_cidr_block.broadcast_address
            ):
                subnet_in_vpc_cidr_range = True
                break

        if not subnet_in_vpc_cidr_range:
            raise InvalidSubnetRangeError(cidr_block)

        # The subnet size must use a /64 prefix length.
        if ipv6_cidr_block and "::/64" not in ipv6_cidr_block:
            raise GenericInvalidParameterValueError("ipv6-cidr-block", ipv6_cidr_block)

        for subnet in self.describe_subnets(filters={"vpc-id": vpc_id}):
            if subnet.cidr.overlaps(subnet_cidr_block):
                raise InvalidSubnetConflictError(cidr_block)

        # if this is the first subnet for an availability zone,
        # consider it the default
        default_for_az = str(availability_zone not in self.subnets).lower()
        map_public_ip_on_launch = default_for_az

        if availability_zone is None and not availability_zone_id:
            availability_zone = "us-east-1a"
        try:
            if availability_zone:
                availability_zone_data = next(
                    zone
                    for zones in RegionsAndZonesBackend.zones.values()
                    for zone in zones
                    if zone.name == availability_zone
                )
            elif availability_zone_id:
                availability_zone_data = next(
                    zone
                    for zones in RegionsAndZonesBackend.zones.values()
                    for zone in zones
                    if zone.zone_id == availability_zone_id
                )

        except StopIteration:
            raise InvalidAvailabilityZoneError(
                availability_zone,
                ", ".join(
                    [
                        zone.name
                        for zones in RegionsAndZonesBackend.zones.values()
                        for zone in zones
                    ]
                ),
            )
        subnet = Subnet(
            self,
            subnet_id,
            vpc_id,
            cidr_block,
            ipv6_cidr_block,
            availability_zone_data,
            default_for_az,
            map_public_ip_on_launch,
            assign_ipv6_address_on_creation=False,
        )

        for tag in tags or []:
            subnet.add_tag(tag["Key"], tag["Value"])

        # AWS associates a new subnet with the default Network ACL
        self.associate_default_network_acl_with_subnet(subnet_id, vpc_id)  # type: ignore[attr-defined]
        self.subnets[availability_zone][subnet_id] = subnet  # type: ignore[index]
        return subnet

    def describe_subnets(
        self, subnet_ids: Optional[List[str]] = None, filters: Optional[Any] = None
    ) -> Iterable[Subnet]:
        # Extract a list of all subnets
        matches = list(
            itertools.chain(*[x.copy().values() for x in self.subnets.copy().values()])
        )
        if subnet_ids:
            matches = [sn for sn in matches if sn.id in subnet_ids]
            if len(subnet_ids) > len(matches):
                unknown_ids = set(subnet_ids) - set(matches)  # type: ignore[arg-type]
                raise InvalidSubnetIdError(list(unknown_ids)[0])
        if filters:
            matches = generic_filter(filters, matches)

        return matches

    def delete_subnet(self, subnet_id: str) -> Subnet:
        for subnets in self.subnets.values():
            if subnet_id in subnets:
                return subnets.pop(subnet_id)
        raise InvalidSubnetIdError(subnet_id)

    def modify_subnet_attribute(
        self, subnet_id: str, attr_name: str, attr_value: str
    ) -> None:
        subnet = self.get_subnet(subnet_id)
        if attr_name in ("map_public_ip_on_launch", "assign_ipv6_address_on_creation"):
            setattr(subnet, attr_name, attr_value)
        else:
            raise InvalidParameterValueError(attr_name)

    def get_subnet_from_ipv6_association(self, association_id: str) -> Optional[Subnet]:
        subnet = None
        for s in self.describe_subnets():
            if association_id in s.ipv6_cidr_block_associations:
                subnet = s
        return subnet

    def associate_subnet_cidr_block(
        self, subnet_id: str, ipv6_cidr_block: str
    ) -> Dict[str, str]:
        subnet = self.get_subnet(subnet_id)
        if not subnet:
            raise InvalidSubnetIdError(subnet_id)
        association = subnet.attach_ipv6_cidr_block_associations(ipv6_cidr_block)
        return association

    def disassociate_subnet_cidr_block(
        self, association_id: str
    ) -> Tuple[str, Dict[str, str]]:
        subnet = self.get_subnet_from_ipv6_association(association_id)
        if not subnet:
            raise InvalidSubnetCidrBlockAssociationID(association_id)
        association = subnet.detach_subnet_cidr_block(association_id)
        return subnet.id, association

    def create_subnet_association(
        self, route_table_id: str, subnet_id: str
    ) -> SubnetRouteTableAssociation:
        subnet_association = SubnetRouteTableAssociation(route_table_id, subnet_id)
        self.subnet_associations[f"{route_table_id}:{subnet_id}"] = subnet_association
        return subnet_association



"""Handles incoming apigatewaymanagementapi requests, invokes methods, returns responses."""
import json
from typing import Any

from moto.core.responses import BaseResponse
from .models import apigatewaymanagementapi_backends, ApiGatewayManagementApiBackend


class ApiGatewayManagementApiResponse(BaseResponse):
    """Handler for ApiGatewayManagementApi requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="apigatewaymanagementapi")

    def setup_class(
        self, request: Any, full_url: str, headers: Any, use_raw_body: bool = False
    ) -> None:
        super().setup_class(request, full_url, headers, use_raw_body=True)

    @property
    def apigatewaymanagementapi_backend(self) -> ApiGatewayManagementApiBackend:
        """Return backend instance specific for this region."""
        return apigatewaymanagementapi_backends[self.current_account][self.region]

    def delete_connection(self) -> str:
        connection_id = self.path.split("/@connections/")[-1]
        self.apigatewaymanagementapi_backend.delete_connection(
            connection_id=connection_id
        )
        return "{}"

    def get_connection(self) -> str:
        connection_id = self.path.split("/@connections/")[-1]
        connection = self.apigatewaymanagementapi_backend.get_connection(
            connection_id=connection_id
        )
        return json.dumps(connection.to_dict())

    def post_to_connection(self) -> str:
        connection_id = self.path.split("/@connections/")[-1]
        data = self.body
        self.apigatewaymanagementapi_backend.post_to_connection(
            data=data,
            connection_id=connection_id,
        )
        return "{}"



"""EMRServerlessBackend class with methods for supported APIs."""
import re
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Iterator
import inspect

from moto.core import BaseBackend, BackendDict, BaseModel
from moto.core.utils import iso_8601_datetime_without_milliseconds
from moto.emrcontainers.utils import get_partition, paginated_list
from .utils import (
    default_auto_start_configuration,
    default_auto_stop_configuration,
    random_appplication_id,
)

from .exceptions import ResourceNotFoundException, ValidationException

APPLICATION_ARN_TEMPLATE = "arn:{partition}:emr-containers:{region}:{account_id}:/applications/{application_id}"

# Defaults used for creating an EMR Serverless application
APPLICATION_STATUS = "STARTED"
JOB_STATUS = "RUNNING"


class FakeApplication(BaseModel):
    def __init__(
        self,
        name: str,
        release_label: str,
        application_type: str,
        client_token: str,
        account_id: str,
        region_name: str,
        initial_capacity: str,
        maximum_capacity: str,
        tags: Dict[str, str],
        auto_start_configuration: str,
        auto_stop_configuration: str,
        network_configuration: str,
    ):
        # Provided parameters
        self.name = name
        self.release_label = release_label
        self.application_type = application_type.capitalize()
        self.client_token = client_token
        self.initial_capacity = initial_capacity
        self.maximum_capacity = maximum_capacity
        self.auto_start_configuration = (
            auto_start_configuration or default_auto_start_configuration()
        )
        self.auto_stop_configuration = (
            auto_stop_configuration or default_auto_stop_configuration()
        )
        self.network_configuration = network_configuration
        self.tags: Dict[str, str] = tags or {}

        # Service-generated-parameters
        self.id = random_appplication_id()
        self.arn = APPLICATION_ARN_TEMPLATE.format(
            partition="aws",
            region=region_name,
            account_id=account_id,
            application_id=self.id,
        )
        self.state = APPLICATION_STATUS
        self.state_details = ""
        self.created_at = iso_8601_datetime_without_milliseconds(
            datetime.today().replace(hour=0, minute=0, second=0, microsecond=0)
        )
        self.updated_at = self.created_at

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        yield "applicationId", self.id
        yield "name", self.name
        yield "arn", self.arn
        yield "autoStartConfig", self.auto_start_configuration,
        yield "autoStopConfig", self.auto_stop_configuration,

    def to_dict(self) -> Dict[str, Any]:
        """
        Dictionary representation of an EMR Serverless Application.
        When used in `list-applications`, capacity, auto-start/stop configs, and tags are not returned. https://docs.aws.amazon.com/emr-serverless/latest/APIReference/API_ListApplications.html
        When used in `get-application`, more details are returned. https://docs.aws.amazon.com/emr-serverless/latest/APIReference/API_GetApplication.html#API_GetApplication_ResponseSyntax
        """
        caller_methods = inspect.stack()[1].function
        caller_methods_type = caller_methods.split("_")[0]

        if caller_methods_type in ["get", "update"]:
            response = {
                "applicationId": self.id,
                "name": self.name,
                "arn": self.arn,
                "releaseLabel": self.release_label,
                "type": self.application_type,
                "state": self.state,
                "stateDetails": self.state_details,
                "createdAt": self.created_at,
                "updatedAt": self.updated_at,
                "autoStartConfiguration": self.auto_start_configuration,
                "autoStopConfiguration": self.auto_stop_configuration,
                "tags": self.tags,
            }
        else:
            response = {
                "id": self.id,
                "name": self.name,
                "arn": self.arn,
                "releaseLabel": self.release_label,
                "type": self.application_type,
                "state": self.state,
                "stateDetails": self.state_details,
                "createdAt": self.created_at,
                "updatedAt": self.updated_at,
            }

        if self.network_configuration:
            response.update({"networkConfiguration": self.network_configuration})
        if self.initial_capacity:
            response.update({"initialCapacity": self.initial_capacity})
        if self.maximum_capacity:
            response.update({"maximumCapacity": self.maximum_capacity})

        return response


class EMRServerlessBackend(BaseBackend):
    """Implementation of EMRServerless APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.region_name = region_name
        self.applications: Dict[str, FakeApplication] = dict()
        self.partition = get_partition(region_name)

    def create_application(
        self,
        name: str,
        release_label: str,
        application_type: str,
        client_token: str,
        initial_capacity: str,
        maximum_capacity: str,
        tags: Dict[str, str],
        auto_start_configuration: str,
        auto_stop_configuration: str,
        network_configuration: str,
    ) -> FakeApplication:

        if application_type not in ["HIVE", "SPARK"]:
            raise ValidationException(f"Unsupported engine {application_type}")

        if not re.match(r"emr-[0-9]{1}\.[0-9]{1,2}\.0(" "|-[0-9]{8})", release_label):
            raise ValidationException(
                f"Type '{application_type}' is not supported for release label '{release_label}' or release label does not exist"
            )

        application = FakeApplication(
            name=name,
            release_label=release_label,
            application_type=application_type,
            account_id=self.account_id,
            region_name=self.region_name,
            client_token=client_token,
            initial_capacity=initial_capacity,
            maximum_capacity=maximum_capacity,
            tags=tags,
            auto_start_configuration=auto_start_configuration,
            auto_stop_configuration=auto_stop_configuration,
            network_configuration=network_configuration,
        )
        self.applications[application.id] = application
        return application

    def delete_application(self, application_id: str) -> None:
        if application_id not in self.applications.keys():
            raise ResourceNotFoundException(application_id)

        if self.applications[application_id].state not in ["CREATED", "STOPPED"]:
            raise ValidationException(
                f"Application {application_id} must be in one of the following statuses [CREATED, STOPPED]. "
                f"Current status: {self.applications[application_id].state}"
            )
        self.applications[application_id].state = "TERMINATED"

    def get_application(self, application_id: str) -> Dict[str, Any]:
        if application_id not in self.applications.keys():
            raise ResourceNotFoundException(application_id)

        return self.applications[application_id].to_dict()

    def list_applications(
        self, next_token: Optional[str], max_results: int, states: Optional[List[str]]
    ) -> Tuple[List[Dict[str, Any]], Optional[str]]:
        applications = [
            application.to_dict() for application in self.applications.values()
        ]
        if states:
            applications = [
                application
                for application in applications
                if application["state"] in states
            ]
        sort_key = "name"
        return paginated_list(applications, sort_key, max_results, next_token)

    def start_application(self, application_id: str) -> None:
        if application_id not in self.applications.keys():
            raise ResourceNotFoundException(application_id)
        self.applications[application_id].state = "STARTED"

    def stop_application(self, application_id: str) -> None:
        if application_id not in self.applications.keys():
            raise ResourceNotFoundException(application_id)
        self.applications[application_id].state = "STOPPED"

    def update_application(
        self,
        application_id: str,
        initial_capacity: Optional[str],
        maximum_capacity: Optional[str],
        auto_start_configuration: Optional[str],
        auto_stop_configuration: Optional[str],
        network_configuration: Optional[str],
    ) -> Dict[str, Any]:
        if application_id not in self.applications.keys():
            raise ResourceNotFoundException(application_id)

        if self.applications[application_id].state not in ["CREATED", "STOPPED"]:
            raise ValidationException(
                f"Application {application_id} must be in one of the following statuses [CREATED, STOPPED]. "
                f"Current status: {self.applications[application_id].state}"
            )

        if initial_capacity:
            self.applications[application_id].initial_capacity = initial_capacity

        if maximum_capacity:
            self.applications[application_id].maximum_capacity = maximum_capacity

        if auto_start_configuration:
            self.applications[
                application_id
            ].auto_start_configuration = auto_start_configuration

        if auto_stop_configuration:
            self.applications[
                application_id
            ].auto_stop_configuration = auto_stop_configuration

        if network_configuration:
            self.applications[
                application_id
            ].network_configuration = network_configuration

        self.applications[
            application_id
        ].updated_at = iso_8601_datetime_without_milliseconds(
            datetime.today().replace(hour=0, minute=0, second=0, microsecond=0)
        )

        return self.applications[application_id].to_dict()


emrserverless_backends = BackendDict(EMRServerlessBackend, "emr-serverless")
